!!     Routines for constructing and manipulating general equilibrium
!!      models implemented with MPSGE
!!      @version 1.0
!!      @author Thomas Rutherford

module mpsge

        implicit none

!       Records of source data:

!!      Defines a line of source code.

        type line
          character (len=2048) :: text	!! Input records are limited to 2048 characters
          integer :: lineno		!! Line number in the input file
          type (line), pointer :: next
        end type line

!!      Defines a list of input lines as they have been read from the input file.

        type linelist
          type (line), pointer :: line
          type (linelist), pointer :: next
        end type linelist

!!      Elements of the symbol table include various MPSGE variables (commodity
!!      prices, sectoral activity level, consumer incomes and auxiliary variables).
!!      The symbol table also includes sets and parameters which are encountered
!!      in the input file.  This list is maintained to provide some context checking
!!      on input.

        type symbol
          character (len=31) :: name		!! Symbol identifier
          character (len=10) :: type		!! One of set, parameter, sector, commodity, consumer, auxiliary
          type (linelist), pointer :: lines     !! Points to lines in which this symbol appears
          type (symbol), pointer :: next        !! Points to next symbol in the global list
          integer :: dim			!! Dimension of the symbol (undefined=-1,scalar=0, vector=1, etc.)
        end type symbol

!!      A domain is simply a list of sets.  This list may include quoted singletons as well as
!!      references to GAMS sets.  Each element of the set is represented by a pointer to the
!!      set in the symbol table

        type domain
          type (symbol), pointer :: s           !! Pointer to this set in the symbol table
          type (domain), pointer :: next        !! Pointer to the next set in this domain (unassociated if no more dimensions)

          integer :: dim			!! Dimension of this domain
        end type domain

!!      An MPSGE input file consists of a series of records.  Each record is composed of
!!      one or more fields.  An input field consist of a label and some field text.

!!      A field in an MPSGE model consists of two primary components: the label and
!!      the field data.  The field data in turn consists of the identifier,
!!      the identifier argument, the spanning sets and the condition clause.

        type field

!! Verbatim text of leading label

          character (len=255)   :: label

!! Verbatim text of following field 

          character (len=255)   :: data

!! Identifier text (data excluding domain and condition)

          character (len=255)   :: identtext

!! Substring text of the identitier argument domain if present

          character (len=255)   :: argtext

!! Substring text of the spanning domain if present (# notation)

          character (len=255)   :: spantext

!! Substring text of the conditional clause if present

          character (len=255)   :: condtn

!! Pointer to field idenfier if the field is readable. 
!! Fields beginning with "(" are skipped.

          type (symbol), pointer :: ident       
                                                

!! Pointer to argument domain if it is readable

          type (domain),pointer :: arg

!! Points to span domain if it is present

          type (domain),pointer :: span         

!! Points to source line

          type (line), pointer :: line          

!! Pointer to next field in the data file.
!! The final field in a datafile is unassociated.

          type (field), pointer :: next
                                                
        end type field

!!      This type describes a _record_ in an MPSGE function, possibly spanning multiple lines
!!      if there are continuations (lines "+" in the first column)

        type record
          type (field), pointer  :: firstfield  !! Points to the first field in this record
          type (line), pointer   :: line        !! Points to the input line where this record begins.
          integer                :: nfield      !! Counts the number of fields in this record
          type (record), pointer :: next        
        end type record

!!      A tax coefficient on a production input or output consists of:
!!      - A tax revenue agent (identified in an A: field)
!!      - An ad-valorem tax rate defined by one of the following:
!!        - A T: field representing a fixed and exogenous tax rate
!!        - A N:/M: field pair representing an endogenous tax.
!!      The tax coefficient consists only of pointers to fields in the input file.  Note that 
!!      the A: field may be on a record which appears prior to the tax rate field.
!!      With endogenous taxes, the N: field defines an auxiliary variable which controls the tax
!!      and the M: field defines a multiplier on the tax rate.

        type taxlist

!!      Pointer to A: field which identifes agent who collects this tax revenue:

              type (field), pointer :: a

!!      Pointer to T: field which defines an exogenous ad-valorem tax rate:

              type (field), pointer :: t

!!      Pointer to N: field indicating the auxiliary variable which determines the
!!      ad-valorem tax rate.  For a given tax entry, either T: or N: will be
!!      used, but not both:

              type (field), pointer :: n

!!      Pointer to an M: field defining the multiplier which is to be applied on 
!!      the previously listed endogenous tax.  An M: field must follow immediately
!!      after an N: field.

              type (field), pointer :: m

!!      Next tax applied on the same production input or output.  This is unassociated
!!      if there are no more applicable taxes:

              type (taxlist), pointer :: next
        end type taxlist

!!      This type defines an input or output in a production block.  The domain associated with
!!      a netput consists of the union of sets in the function declaration block and the 
!!      commodity price.  For example, if we have
!!>
!!      $prod:y(i,t)
!!              o:p(j,t)        q:y0(i)
!!              ...
!!<
!!      then the _domain_ of the netput, _d_, is (_i,t,j_), and the _subdomain_ of the netput, _sd_, is (_j_).
!!      

        type netput
                type (field), pointer    :: f           !! Pointer to the I: or O: field
                type (field), pointer    :: qref        !! Pointer to the associated Q: field in the same record
                type (field), pointer    :: pref        !! Pointer to the associated P: field in the same record
                type (domain), pointer   :: d           !! Pointer to the domain of this netput
                type (domain), pointer   :: sd          !! Pointer to the subdomain of the netput
                type (taxlist), pointer  :: t           !! Pointer to the taxes which apply to this netput
		character (len=1024) :: nestcount
		integer :: nestfields
		logical :: output			!! Logical flag for an output coefficient
                integer :: index                        !! Integer index associated with this netput
        end type netput


!!      List of netputs references from the associated nest.

        type netputlist
                type (netput), pointer :: np            !! Netput at this location in the list
                type (field),  pointer :: f             !! Nest assignment field
                type (netputlist), pointer :: next      !! Pointer to next element in list (nul if last input)
        end type netputlist

!!      Nest type characterizes a CES nest in a cost function.

        type nest
                character (len=255) :: ident		!! Identifier text for this nest 
                character (len=255) :: parentid		!! Identifier text for the parent nest
                character (len=255) :: value            !! Text field of the elasticity value
                logical :: output                       !! Logical flag identifies type of nest (output or input)
                type (field), pointer :: f              !! Pointer to the input field which defines the nest
                type (nest), pointer :: parent          !! Pointer to parent nest
                type (netputlist), pointer :: elements  !! Pointer to a list of constituent inputs or outputs
                type (nestlist), pointer :: subnests    !! Pointer to a list of constituent nests
                type (nest), pointer :: next            !! Pointer to next nest in current function
                integer :: index                        !! Integer index associated with this nest
		integer :: level			!! Level in the nesting structure (top level = 0)
        end type nest

!!      Linked list of nests used to represent the nested CES substitution structure:

        type nestlist
                type (nest), pointer :: n               !! Pointer to a nest in this list
                type (nestlist), pointer :: next        !! Pointer to next element of the list
        end type nestlist

!! MPSGE function ($prod, $demand or $report)

        type function           
          type (record), pointer :: firstrecord !! Pointer to first input record in this function
          integer               :: nrecord      !! Count of number of records in this function
          integer               :: nfield       !! Count of number of fields in this function
          type (nest), pointer  :: inputs       !! Points to the top level input nest    
          type (nest), pointer  :: outputs      !! Points to the top level output nest    
          type (function), pointer :: next      !! Pointer in linked list of all functions
	  integer		:: index	!! Index counter for associated names
        end type function

        type (line), pointer     :: inputfile   !! Pointer to first line in the linked list of input file lines
        type (function), pointer :: functionlist !! Pointer to the first function in linked list
        type (symbol), pointer   :: idlist      !! Pointer to the first symbol
        type (symbol), pointer   :: idlast      !! Pointer to the last symbol

        integer :: nfun=0               !! Number of functions in the MPSGE file
        integer, parameter :: iogms=10  !! Unit number of the MPSGE file

	logical :: splitlog=.false.	!! Flag for log output of field parsing operation
	logical :: log=.false.		!! Flag for general log output


contains

!!      Master routine to read an MPSGE input file:
!!        - Open the file
!!        - Read one line at a time into a linked list for listing.
!!        - Locate functions and set up pointers to the constituent records.
!!        - Parse records into fields.
!!
!!	None of the logical structure of the underlying economic model is
!!	formulated at this stage.

        subroutine readmps
        integer :: lineno
	type (line), pointer :: L
        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f, ff
        character (len=2048) :: input=' '

        nullify(inputfile)
	nullify(functionlist)
	nullify(f)
	nullify(L)

	write(*,'(a)') '$ontext'

        lineno = 0
        do

!       Read a line until we reach one which is not a comment:
!	L is nullified at end of file.

	  call readline(L, input, lineno)
	  if (.not.associated(L)) exit

!       Read a new function?

          if (input(1:1).eq.'$') then
            nfun = nfun + 1
            if (associated(functionlist)) then
	      call newfunction(fn%next); fn => fn%next
            else
	      call newfunction(functionlist); fn => functionlist
            endif
          else
            if (.not.associated(functionlist)) &
              stop 'Error == MPSGE input must begin with a function.'
          end if

!       Read a new record?

          if (input(1:1).ne.'+') then
            if (associated(fn%firstrecord)) then
	      call newrecord(r%next);  r => r%next; 
            else
	      call newrecord(r); fn%firstrecord => r; 
            end if
            r%line => L
            fn%nrecord = fn%nrecord + 1
          end if
	  nullify(f)

!	Keep pulling fields off the input record until there are no more:

          do while (input.ne.' ')

	    call newfield(ff)

!	    if (associated(lastfield)) lastfield%next => ff

!	This logic only applies to the first field extracted from the
!	line:

            if (input(1:1).ne.'+') r%firstfield => ff
	    if (associated(f)) ff%next => f

            f => ff
            f%line => L
            r%nfield = r%nfield + 1
	    fn%nfield = fn%nfield + 1

	    call readfield(input,f)
          end do

        end do
	write(*,'(a)') '$offtext'
	return
        end subroutine readmps

!!	Read to the next non-blank line from the input file, incrementing
!!	the line count:

	subroutine readline(L, input,lineno)
	type (line), pointer :: L
        character (len=*) :: input
	integer :: lineno
	integer :: ic

	input = ' '
	do while (input.eq.' ') 
	  read(iogms,'(a)',end=100) input
	  lineno = lineno + 1
	  write(*,'(a,i0,a,t12,a)') '[',lineno,']',trim(input)
            
	  if (associated(inputfile)) then
	    call newline(L%next)
	    L => L%next
	  else
	    call newline(inputfile)
	    L => inputfile
	  endif
	  L%text = input
	  L%lineno = lineno
            
	  ic = index(input,'!')
	  if (ic.gt.0) input(ic:) = ' '
	  call tab2blank(input)       ! Convert tabs to blanks:
	  if (input(1:1).eq.'*') input = ' '
	end do
	return
100	nullify(L)
	return
	end subroutine readline

!!	Read the last field from a line of input, using the 
!!	colon to locate the label and the data components:	

	subroutine readfield(input,f)
	type (field), pointer :: f
        character (len=*) :: input
	integer :: ic

	ic = index(input,':',back=.true.)
	if (ic.eq.0) stop 'No colon?'
	f%data = adjustl(input(ic+1:))
	input(ic:) = ' '
	ic = index(input(1:ic-1),' ',back=.true.)
	f%label = input(ic+1:)
	call lower(f%label)
	input(ic+1:) = ' '
	if (input.eq. '+') input = ' '
	return
	end subroutine readfield

!!      Fold a character string to lower case.

        subroutine lower(word)
        IMPLICIT NONE
        character (len=*) :: word  !! Character to be placed in lower case.
        integer :: i
        do 100 i=1,len(word)
          if (word (i:i).eq.'A') word (i:i) = 'a'
          if (word (i:i).eq.'B') word (i:i) = 'b'
          if (word (i:i).eq.'C') word (i:i) = 'c'
          if (word (i:i).eq.'D') word (i:i) = 'd'
          if (word (i:i).eq.'E') word (i:i) = 'e'
          if (word (i:i).eq.'F') word (i:i) = 'f'
          if (word (i:i).eq.'G') word (i:i) = 'g'
          if (word (i:i).eq.'H') word (i:i) = 'h'
          if (word (i:i).eq.'I') word (i:i) = 'i'
          if (word (i:i).eq.'J') word (i:i) = 'j'
          if (word (i:i).eq.'K') word (i:i) = 'k'
          if (word (i:i).eq.'L') word (i:i) = 'l'
          if (word (i:i).eq.'M') word (i:i) = 'm'
          if (word (i:i).eq.'N') word (i:i) = 'n'
          if (word (i:i).eq.'O') word (i:i) = 'o'
          if (word (i:i).eq.'P') word (i:i) = 'p'
          if (word (i:i).eq.'Q') word (i:i) = 'q'
          if (word (i:i).eq.'R') word (i:i) = 'r'
          if (word (i:i).eq.'S') word (i:i) = 's'
          if (word (i:i).eq.'T') word (i:i) = 't'
          if (word (i:i).eq.'U') word (i:i) = 'u'
          if (word (i:i).eq.'V') word (i:i) = 'v'
          if (word (i:i).eq.'W') word (i:i) = 'w'
          if (word (i:i).eq.'X') word (i:i) = 'x'
          if (word (i:i).eq.'Y') word (i:i) = 'y'
          if (word (i:i).eq.'Z') word (i:i) = 'z'
          if (word (i:i).eq.char(9)) word(i:i) = ' '
100     continue
        return
        end subroutine lower


!!      Convert tabs on an input line to spaces:

        subroutine tab2blank(line)
        character (len=*) :: line  !! Line to scan for tabs
        character :: tab=char(9)
        integer :: k
        do k=1,len_trim(line)
          if (line(k:k).eq.tab) line(k:k) = ' '
        end do
        return
        end subroutine tab2blank

!!      Evaluate the dimension of a domain by counting
!!      the number of set identifiers in the linked list
!!      defining the domain:

        integer function dimension(d)
        type (domain), pointer :: d     !! Domain to be evaluated
        type (domain), pointer :: dd    
        dimension = 0
        if (.not.associated(d)) return
        dd => d
        do while (associated(dd))
          dimension = dimension + 1
          dd => dd%next
        end do
        end function dimension

!!      Generate an echoprint of the file.  This is currently very preliminary,
!!      and will later be recoded to produce a more carefully formatted report.

        subroutine echoprint

        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f
	integer :: i

        fn => functionlist
        nullify(f);
        do while (associated(fn))
          write(*,'(a)') ' ===== '
	  r => fn%firstrecord
	  do while (associated(r))
	    f => r%firstfield
	    do while (associated(f))
	      write(*,'(a)') ' '//trim(f%label)//':'//trim(f%data)
	      f => f%next
	    end do
	    r => r%next
	  end do
	  fn => fn%next
        end do

	return
	end subroutine echoprint

!!      Parse all fields in the file into constituent elements:
!!      - Field identifier text
!!      - Identifier argument text
!!      - Spanning domain text
!!      - Condition text.

        subroutine parse
        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f

        fn => functionlist
        nullify(f);
        do while (associated(fn))
	  r => fn%firstrecord
	  do while (associated(r))
	    f => r%firstfield
	    do while (associated(f))

!	Sort out the data:

	      call split(f%data, f%identtext, f%argtext, f%spantext, f%condtn)

!       Read domain for each field:

	      if (f%argtext .ne.' ') call readdomain(f%argtext,  f%arg, f%line)
	      if (f%spantext.ne.' ') call readdomain(f%spantext, f%span, f%line)
	      f => f%next
	    end do
	    r => r%next
	  end do
	  fn => fn%next
	end do
	return
        end subroutine parse

!! Return pointer to given identifier

        subroutine findid(name, id)             
        character (len=*), intent(in) :: name   !! Identifier for the symbol to be located
        type (symbol), pointer :: id            !! Pointer into the master list for the located symbol.
        type (symbol), pointer :: s

        nullify(id)
        s => idlist
        do while (associated(s))
          if (s%name==name) then
            id => s
            return
          end if
          s => s%next
        end do
        return 
        end subroutine findid

!! Introduce a new identifier in the master list.

        subroutine addid(name, type, s, L)      

!!      Name of the identifier to be added to the master list

        character (len=*), intent(in) :: name

!!      Description of the type of item to be introduced:
!!      - _set_: a GAMS set identifier
!!      - _parameter_ : a GAMS parameter identifier
!!      - _sector_ : an MPSGE sector (acivity variable)
!!      - _commodity_ : an MPSGE commodity (price variable)
!!      - _consumer_ : an MPSGE consumer (income variable)
!!      - _auxiliary_ : an MPSGE auxiliary variable

        character (len=*), intent(in) :: type
        type (symbol), pointer :: s
        type (line), pointer :: L
        if (.not.associated(idlist)) then
	  call newsymbol(idlist)
          s => idlist
        else
	  call newsymbol(idlast%next)
          s => idlast%next
        end if
        s%name = name
        s%type = type
        allocate(s%lines);
        s%lines%line => L
        nullify(s%lines%next)
        nullify(s%next)
        idlast => s
        return
        end subroutine addid

!	Create a copy of a domain.

        subroutine domaincopy(d1,d2)            
        type (domain), pointer:: d1	!! Domain to be copied
        type (domain), pointer :: d2	!! New domain to be created as a copy of d1
        type (domain), pointer :: d,dd
        if (.not.associated(d1)) return

	call newdomain(d2)
        d  => d1; dd => d2
        do while (associated(d))
          dd%s => d%s
          if (associated(d%next)) then
	    call newdomain(dd%next); dd => dd%next
          endif
          d => d%next
        end do
        return
        end subroutine domaincopy

!!	Add an entry in the linked list of line numbers which refer to a given symbol.

        subroutine addlineref(L,s)

        type (symbol), pointer  :: s  !! Symbol which is referenced
        type (line), pointer :: L	!! Line in which the symbol is referenced
        type (linelist), pointer :: LL

        LL => s%lines
        do while (associated(LL))
          if (L%lineno.eq.LL%line%lineno) return
          if (.not. associated(LL%next)) exit
          LL => LL%next
        end do
        allocate(LL%next)
        LL => LL%next
        LL%line => L
        nullify(LL%next)
        return
        end subroutine addlineref

!! Add a set element to the domain

        subroutine domainadd(name,dm, L)     
        character (len=*), intent (in) :: name	!! Identier of set to be added to the domain
        type (domain), pointer :: dm  !! Domain to which this set is to be introduced
        type (line), pointer :: L  !! Line in which the set reference was made.

        type (symbol), pointer  :: s
        type (domain), pointer :: d

!       Find what set we are introducing

        call findid(name, s)
        if (associated(s)) then
          if (s%type.ne.'set') stop 'Error -- not a SET'
          call addlineref(L, s)
        else
          call addid(name,'set',s,L)
        end if

        if (.not. associated(dm)) then
	  call newdomain(dm); dm%s => s; 
          return
        endif
        
!       Add this set at the end of the domain list:

        d => dm
        do 
          if (.not.associated(d%next)) exit
          d => d%next
        end do
	call newdomain(d%next)
	d=>d%next; d%s => s; 

        return
        end subroutine domainadd


!!  Logical function which reports whether a set with a given name
!!  is in a domain:

        logical function indomain(s,dm)
        type (symbol), pointer	:: s	!! Set to be located
        type (domain), pointer	:: dm	!! Domain in which the set is to be located.

        type (domain), pointer :: d

        if (.not. associated(dm)) then
          indomain =.false.
          return
        endif

        d => dm
        do while (associated(d))
          if (associated(d%s,s)) then
            indomain = .true.
            return
          endif
          d => d%next
        end do
        indomain = .false.
        return
        end function indomain

!!   Tets whether one domain (_d1_) is a subst of another (_d2_).

        logical function subset(d1,d2) 
        type (domain), pointer :: d1	!! Subset domain
	type (domain), pointer :: d2	!! Superset domain 
	type (domain), pointer :: d, dd
        d  => d1; 
        do while (associated(d))
          dd => d2;
          subset = .false.
          do while (associated(dd))
            if (d%s%name.eq.dd%s%name) then
              subset = .true.
              exit
            end if
            dd => dd%next
          end do
          if (.not.subset) return
          d => d%next
        end do
        return
        end function subset

!!      Split up an input field into the constituent elements
!!      - Field identifier text
!!      - Identifier argument text
!!      - Spanning domain text
!!      - Condition text.
!!
!!      For example:
!!>
!!      i:p(i,t)#(s)$beta       
!!<
!!      parses into:
!!>
!!      Field identifier: p
!!      Argument text: i,t
!!      Spanning domain text: s
!!      Condition text: beta
!!<
        subroutine split(field,ident,arg,span,condtn)
        character (len=*), intent(in)  :: field
        character (len=*), intent(out) :: ident,arg,span,condtn
        integer :: ic, np

        ident = ' '
        arg = ' '
        span = ' '
        condtn = ' '

!       Split field text into text fields for the identifier, the argument domain,
!       the span domain and conditional clause:

        np = 0
        ident = field
        do ic=1,len_trim(field)
          if (field(ic:ic).eq.'(') np = np + 1
          if (field(ic:ic).eq.')') np = np - 1
          if (ic.gt.1 .and. field(ic:ic).eq.'$' .and. np.eq.0) then
            condtn = field(ic+1:)
            ident = field(1:ic-1)
            exit
          end if
        end do
        if (np.ne.0) stop ' *** Parethesis mismatch.'

        ic = index(ident,'#')
        if (ic.gt.0) then
          span = adjustl(ident(ic+1:))
          ident(ic:) = ' '
        end if

        ic = index(ident,'(')
        if (ic.gt.1) then
          arg = ident(ic:)
          ident(ic:) = ' '
        end if

	if (splitlog) then
	  write(*,'(a)') ' --------------'
	  write(*,'(a)') ' field  = '//trim(field)
	  write(*,'(a)') ' ident  = '//trim(ident)
	  write(*,'(a)') ' arg    = '//trim(arg)
	  write(*,'(a)') ' span   = '//trim(span)
	  write(*,'(a)') ' condtn = '//trim(condtn)
	end if

        return
        end subroutine split

!!	Read a domain and parse into a series of set elements.  

        subroutine readdomain(dtext,d,L)
        character (len=*), intent(in) :: dtext	!! Domain to be parsed
        type (domain), pointer :: d  !! Pointer to the domain list
        type (line), pointer :: L  !! Pointer to source line
        character (len=255) :: textfield
        character (len=31) :: setid
        integer :: ic

        textfield = dtext

!       Check for leading and closing parentheses (required):

        if (len_trim(textfield).le.2) then
          stop ' *** READDOMAIN: invalid domain text.'
        endif

        if (textfield(1:1).ne.'(') then
          stop ' *** READDOMAIN: opening parenthes in domain text is missing.'
        endif

        if (textfield(len_trim(textfield):len_trim(textfield)).ne.')') then
          stop ' *** READDOMAIN: closing parenthesis in domain text is missing.'
        endif

        textfield = adjustl(textfield(2:len_trim(textfield)-1))

!       Check that there are no extra paretheses:

        if (index(textfield,'(') + index(textfield,')')  + &
            index(textfield,';') + index(textfield,':').gt.0) then
          stop ' *** READDOMAIN: embedded parenthesese or semicolons in domain text.'
        endif

        call lower(textfield)
        do while (textfield.ne.' ')
          ic = index(textfield,',')
          if (ic.eq.0) then
            setid = adjustl(textfield)
            textfield = ' '
          else
            if (ic.eq.1) stop ' *** READDOMAIN: empty set specification.'
            setid = adjustl(textfield(1:ic-1))
            textfield = textfield(ic+1:)
          end if
          call domainadd(setid, d, L)
        end do

        return
        end subroutine readdomain

!!	Allocate a new line and assign default values

	subroutine newline(L)
	type (line), pointer :: L
	nullify(L)
	allocate(L)
	nullify(L%next)
	L%lineno = 0
	L%text = ' '
	return
	end subroutine newline

!!	Allocate a new symbol and assign default values
!!	for all the symbol characteristics

	subroutine newsymbol(s)
	type (symbol), pointer :: s
	nullify(s)
	allocate(s)
	s%name = ' '
	s%type = 'undefined'
	nullify(s%lines)
	nullify(s%next)
	s%dim = -1
	return
	end subroutine newsymbol

!!	Allocate a new function pointer with default values

	subroutine newfunction(fn)
	type (function), pointer :: fn
	nullify(fn)
	allocate(fn)
	fn%nfield = 0
	fn%nrecord = 0
	nullify(fn%next)
	nullify(fn%inputs)
	nullify(fn%outputs)
	nullify(fn%firstrecord)
	fn%index = 0
	return
	end subroutine newfunction


!!	Allocate a new domain pointer with default values

	subroutine newdomain(d)
	type (domain), pointer :: d
	nullify(d)
	allocate(d)
	nullify(d%next)
	nullify(d%s)
	d%dim = -1
	return
	end subroutine newdomain

!!	Allocate a new field with default values

	subroutine newfield(f)
	type (field), pointer :: f
	nullify(f)
	allocate(f)
	f%label = ' '
	f%data = ' '
	f%identtext = ' '
	f%argtext = ' '
	f%spantext = ' '
	f%condtn = ' '
	nullify(f%ident)
	nullify(f%arg)
	nullify(f%span)
	nullify(f%line)
	nullify(f%next)
	return
	end subroutine newfield

!!	Allocate a new record with default values

	subroutine newrecord(r)
	type (record), pointer :: r
	nullify(r)
	allocate(r)
	nullify(r%firstfield)
	nullify(r%line)
	r%nfield = 0
	nullify(r%next)
	return
	end subroutine newrecord


!!	Allocate a new tax record with default values

	subroutine newtax(t)
	type (taxlist), pointer :: t
	nullify(t)
	allocate(t)
	nullify(t%a)
	nullify(t%t)
	nullify(t%n)
	nullify(t%m)
	nullify(t%next)
	return
	end subroutine newtax

!!	Allocate a new netput with default values

	subroutine newnetput(np)
	type (netput), pointer :: np
	nullify(np)
	allocate(np)
	nullify(np%f)
	nullify(np%qref)
	nullify(np%pref)
	nullify(np%d)
	nullify(np%sd)
	nullify(np%t)
	np%nestcount = ' '
	np%nestfields = 0
	np%index = 0
	return
	end subroutine newnetput

!!	Allocate a new nest with default values

	subroutine newnest(n)
	type (nest), pointer :: n
	nullify(n)
	allocate(n)
	n%ident = ' '
	n%parentid = ' '
	n%value = ' '
	n%output = .false.
	nullify(n%f)
	nullify(n%next)
	nullify(n%parent)
	nullify(n%elements)
	nullify(n%subnests)
	n%index = 0
	n%level = 0
	return
	end subroutine newnest

!!	Add d1 to union, excluding quoted singletons

	subroutine union(d1,u)
	type (domain), pointer :: d1, u, d, dd

	if (.not.associated(d1)) stop 'd1 is not associated'
	if (.not.associated(u)) nullify(u)

	d => d1 
	nullify(dd)
	if (associated(u)) dd => u
	do while (associated(d))
	  if (.not.singleton(d%s%name) .and. .not.indomain(d%s,u)) then
	    if (associated(u)) then
	      call newdomain(dd%next)
	      dd => dd%next
	    else
	      call newdomain(u)
	      nullify(dd)
	      dd => u
	    end if
	    dd%s => d%s
	  end if
	  d => d%next
	end do
	return
	end subroutine union

!!	Determine the subdomain of d1 in d2, i.e. those element of
!!	d1 which are not in d2:
!!>
!!                  subdomain = d1 \ (d1 ^ d2)
!!<
!!	The output is the set of elements in d1 which are not 
!!	singletons and not in d2.
!!
	subroutine findsubdomain(d1,d2,subdomain)
	type (domain), pointer :: d1, d2, subdomain, d, dd

	nullify(subdomain)

!	Look at each element in d1:

	d => d1 
	do while (associated(d))
	  if (.not.indomain(d%s,d2)) then
	    if (associated(subdomain)) then
	      call newdomain(dd%next)
	      dd => dd%next
	    else
	      call newdomain(subdomain)
	      dd => subdomain
	    end if
	    dd%s => d%s
	  end if
	  d => d%next
	end do
	return
	end subroutine findsubdomain

	logical function singleton(string)
	character (len=*) :: string
	singleton = ((string(1:1).eq.'"') .or. (string(1:1).eq."'"))
	return
	end function singleton

	subroutine writeparam(ident, index, d, string)
	character (len=*) :: ident, string
	type (domain), pointer :: d
	integer :: index
	character (len=255) :: domstr

	call writedomain(d, domstr)
	if (domstr.eq.' ') then
	  write(string,'(a,i3.3)') trim(ident), index
	else
	  write(string,'(a,i3.3,a)') trim(ident), index,'('//trim(domstr)//')'
	end if
	return
	end subroutine writeparam

	subroutine writedomain(d, string)
	character (len=*) :: string
	type (domain), pointer :: d, dd
	dd => d
	string = ' '
	do while (associated(dd))
	  if (string.eq.' ') then
	    string = dd%s%name
	  else
	    string = trim(string)//','//dd%s%name
	  end if
	  dd => dd%next
	end do
	return
	end subroutine writedomain

!!	Read through records for each of the functions and add symbols
!!	to the master list as they are encountered.  Check that the 
!!	input data conforms to syntax rules.

	subroutine chkprod

	type (symbol), pointer :: s
	type (function), pointer :: fn
	type (field), pointer :: f
	type (record), pointer :: r
	logical :: pfield
	integer :: fieldno


	fn => functionlist

	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  if (f%label.ne.'$prod') then
	    fn => fn%next
	    cycle 
	  end if
	  if (log) write(*,'(a)') ' ** '//trim(f%data)

!	Skip the elasticity nests on the first pass.


	  do while (associated(r%next))

	    r => r%next
	    pfield = .false.	! Flag to assure only one P: field per record
	    f => r%firstfield
	    fieldno = 1

	    do while (associated(f))

	      if (log) write(*,'(a)') ' -- '//trim(f%label)//':'//trim(f%data)

	      call chkfield(f,fieldno,pfield)

	      fieldno = fieldno + 1
	      f => f%next
	    end do
	  end do
	  fn => fn%next
	end do
	return
	end subroutine chkprod

!!	Evaluate fields in a production block record:
!!	- identifier dimension and type must be consistent
!!	- i: and o: must only appear as leading fields and reference prices
!!	- q: field must be second and reference a parameter 
!!	- p: fields may appear only once and must reference parameter
!!	- a: field must reference a consumer
!!	- n: field must reference an auxiliary variable 
!!	- m: field must reference a parameter
!!	- t: field must reference a parameter
!!
	subroutine chkfield(f,fieldno,pfield)
	integer :: fieldno
	logical :: pfield
	type (field), pointer :: f
	type (symbol), pointer :: s

!	We are scanning the fields assoiciated with named labels, hence we can skip
!	the nest-assignment fields:

	if (f%data.eq.' ' .or. f%data(1:1).eq.'$') then
	  if (index(f%label,'.tl').gt.0) stop ' *** Error: .tl syntax is not yet implemented.'
	  if (index(f%label,'.te').gt.0) stop ' *** Error: .te syntax is not supported.'
	  return
	endif

	select case (f%label)	! Evaluate fields which are not nesting assignments

	case('i')
	  if (fieldno.ne.1) stop ' *** Error: I: must be the leading field.'
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'commodity') stop ' *** Error: I: field must reference a commodity price'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of commodity price is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'commodity', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('o')
	  if (fieldno.ne.1) stop ' *** Error: O: must be the leading field.'
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'commodity') stop ' *** Error: O: field must reference a commodity price'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of commodity price is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'commodity', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('q')
	  if (fieldno.ne.2) stop ' *** Error: Q: must be in the second field.'
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'parameter') stop ' *** Error: Q: field must reference a parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('p')
	  if (pfield) stop ' *** Error: P: field appears more than once.'
	  pfield = .true.
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'parameter') stop ' *** Error: P: field must reference a parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('a')
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	    if (s%type.ne.'consumer') stop ' *** Error: A: field must reference a consumer'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'consumer', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if


	case('n')
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	    if (s%type.ne.'auxiliary') stop ' *** Error: N: field must reference an AUXILIARY'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of AUXILIARY variable changes.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'auxiliary', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('m')
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	  if (s%type.ne.'parameter') stop ' *** Error: M: multiplier must reference an parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter variable changes.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('t')
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	    if (s%type.ne.'parameter') stop ' *** Error: T: multiplier must reference an parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter variable changes.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case default
	  stop ' *** Error: unrecognized field label in $prod block'

	end select 

	return

	end subroutine chkfield

!!	Read nest assignments from the first record of each $prod block

	subroutine readnest

	type(function), pointer :: fn
	type(record), pointer :: r
	type(field), pointer :: f
	type(nest), pointer ::	esub, &	! Top level input nest
				etrn, &	! Top level output nest
				n, &	! Last nest allocated
				nn,np	! Nest pointer

	logical :: subnest	! Logical flag that a subnest has been specified.
	integer :: fieldno
	character (len=255) :: nestid

!	First pass to set up nesting pointers:

	fn => functionlist
	nullify(f)
	nullify(n)
	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield

!	Set up the top level input and output nests:

	  call newnest(esub)
	  fn%inputs  => esub
	  esub%value = '0'
	  esub%ident = 's'

	  call newnest(etrn)
	  fn%outputs => etrn
	  etrn%value = '0'
	  etrn%ident = 't'
	  etrn%output = .true.

	  esub%next => etrn
	  n => etrn
	  subnest = .false.
	  fieldno = 1

!	Look through all the fields on this record following the first:

	  do while (associated(f%next))
	    f => f%next
	    fieldno = fieldno + 1
	    if (f%spantext.ne.' ') stop ' *** Span notation "#" not permited in elasticity.'

!	If an elasticity is specified as a parameter, verify that the parameter domain
!	is a subset of the function domain:

	    if ((f%argtext.ne.' ') .and. (.not.subset(f%arg, r%firstfield%arg))) then
    	      write(*,'(a)') ' '//trim(r%firstfield%label)//':'//trim(r%firstfield%data)//' => '//trim(r%firstfield%argtext)
    	      write(*,'(a)') ' '//trim(f%label)//':'//trim(f%data)//' => '//trim(f%argtext)
	      stop ' ***  Invalid elasticity field domain.'
	    end if

!	Nest ID is specfied in the field label:

    	    nestid = f%label
	    if (index(nestid,'.tl').gt.0) stop ' *** Error: .tl syntax is not yet implemented.'
	    if (index(nestid,'.te').gt.0) stop ' *** Error: .te syntax is not supported.'

	    if (nestid.eq.'s') then
	      if (subnest) stop ' *** s: field must preceed subnest fields.'
	      if (associated(esub%f)) stop ' *** Multiple s: fields in $prod block'
	      if (fieldno.gt.3)       stop ' *** s: field must be first or second elasticity'
	      esub%f => f
	      esub%value = f%data
	      cycle
	    endif

	    if (nestid.eq.'t') then
	      if (subnest) stop ' *** t: field must preceed subnest fields.'
	      if (associated(etrn%f)) stop ' *** Multiple t: fields in $prod block'
	      if (fieldno.gt.3)       stop ' *** t: field must be first or second elasticity'
	      etrn%f => f
	      etrn%value = f%data
	      cycle
	    endif

	    subnest = .true.
	    call newnest(n%next)
	    n => n%next

	    call parsenestid(nestid,n%ident,n%parentid)
	    n%f => f
	    n%value = f%data
	  end do

!	Now install pointers to parents where an explicit parent has
!	been declared:

	  n => fn%inputs
	  do while (associated(n))
	    if (n%parentid.eq.' ') then
	      if (.not.(associated(n,fn%inputs) .or. associated(n,fn%outputs))) n%level = 1
	      n => n%next
	      cycle
	    end if

!	An explicit parent has been declared - find it

	    call findnest(n%parentid, fn, np)
	    n%parent => np
	    n%level = np%level + 1

!	Add this nest to the list of subnests in p:

	    call addsubnest(n, np)
	    n => n%next
	  end do 
	    
	  if (log) then
	    write(*,'(a)') ' Nests:'
	    n => fn%inputs
	    do while (associated(n))
	      if (associated(n%parent)) write(*,'(a)') ' '//trim(n%ident)//' -> '//trim(n%parentid)
	      n => n%next
	    end do
	  end if

	  fn => fn%next
	end do

	return
	end subroutine readnest

!!	Split a nest elasticity field into a nest identifier and a 
!!	parent identifier:

	subroutine parsenestid(nestid,ident,parentid)
	character (len=*) :: nestid	!! Identifier for nest
	character (len=*) :: ident	!! Identifier of the nest itself
	character (len=*) :: parentid	!! Identifier of the nest parent 

	integer :: ic, jc
	ic = index(nestid,'(')
	jc = index(nestid,')',back=.true.)
	if (ic.gt.0 .and. jc.eq.0) then
	  write(*,'(a)') trim(nestid)
	  stop ' Invalid parent node -- missing closing parenthesis.'
	end if
	if (ic.eq.0 .and. jc.gt.0) then
	  write(*,'(a)') trim(nestid)
	  stop ' Invalid parent node -- missing opening parenthesis.'
	end if
	if (ic.eq.0) then
	  ident = nestid
	  parentid = ' '
	else
	  if (jc.lt.ic+2) then
	    write(*,'(a)') trim(nestid)
	    stop ' Invalid parent node -- parenthesis mismatch.'
	  end if
	  ident = nestid(1:ic-1)
	  parentid = nestid(ic+1:jc-1)
	end if
	return
	end subroutine parsenestid

!!	Read through the production function coefficients:
!!	- Create netput descriptions of each production input and output
!!	- Create linked list data structure characterizing the nested CES function

	subroutine readcoef
	type (function), pointer :: fn
	type (symbol), pointer :: s
	type (record), pointer :: r
	type (netput), pointer :: np
	type (domain), pointer :: d
	type (field), pointer :: f, fnf, taxagent, lastfield
	type (nest), pointer :: n
	logical :: input, conditional
	integer :: nparent

	nullify(taxagent)	! No tax agent defined at this point
	  
	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  fnf => r%firstfield
	  if (fnf%label.ne.'$prod') then
	    fn => fn%next
	    cycle
	  end if

!	Read all the records in this function after the elasticity record:

	  do while (associated(r%next))
	    r => r%next

!	Look through the fields in this record:

	    f => r%firstfield
	    nparent = 0
	    call findid(f%identtext, s)

!	Create a netput for this record:

	    call newnetput(np)
	    np%f => f
	    np%output = (f%label.eq.'o')

	    np%nestcount = '0'
	    np%nestfields = 0
	    conditional = .true.

	    if (associated(fnf%arg)) call union(fnf%arg, np%d)
	    if (associated(f%arg))   call union(f%arg,  np%d)
	    if (associated(f%span))  call union(f%span, np%d)
	    if (associated(np%d))    call findsubdomain(np%d, fnf%arg, np%sd)
	    nullify(lastfield)	! Reading the first field -- no prior field at this point


!	Look for nest identifiers:

	    do while (associated(f%next))
	      f => f%next

!	A nest identifier is a field with no data component:

	      if (f%data.eq.' ') then

!	Locate this nest identifier in the master list:

	        call findnest(f%label, fn, n)

!	Record the nest type:

		call nesttype(fn, n, np%output)

!	Add this netput to this nest:

		call addnetput(np, n, f)
		nullify(lastfield)
		np%nestfields = np%nestfields + 1
		if (f%condtn.ne.' ') then
		  np%nestcount = trim(np%nestcount)//'+1$'//trim(f%condtn)
		else
		  np%nestcount = trim(np%nestcount)//'+1'
		  conditional = .false.
		end if
	      else
	        call readelem(np, f, taxagent, lastfield)
		lastfield => f
	      end if
	    end do

!	Assign this netput to a toplevel

	    if (     np%output) n => fn%outputs
	    if (.not.np%output) n => fn%inputs
	    if (np%nestfields.eq.0) then
	      nullify(f)
	      call addnetput(np, n, f)
	    else
	      if (conditional) then
	        call newfield(f)
		f%condtn = '('//trim(np%nestcount)//'=0)'
		call addnetput(np, n, f)
	      endif
	    end if
	  end do
	  fn => fn%next
	end do

	return
	end subroutine readcoef

!!	Introduce type definitions for nests beginning at n
!!	and proceeding to the top of the tree.

	subroutine nesttype(fn, n, output)
	type (function), pointer :: fn	!! Function we fare looking at
	type (nest), pointer :: n	!! Nest from which to trace
	logical :: output		!! Flag for whether this is an output nest
	type (nest), pointer :: np
	if (n%output .and. .not.output) stop 'Error: input enters an output nest'
	n%output = output
	if (.not. associated(n%parent)) then
	  if (output) then
	    n%parent => fn%outputs
	  else
	    n%parent => fn%inputs
	  end if
	  return
	else
	  np => n
	  do while (associated(np%parent))
	    np => np%parent
	    if (np%output .and. .not.output) stop 'Error: input enters an output nest'
	    np%output = output
	  end do
	  if (output      .and. .not.associated(np,fn%outputs)) np%parent => fn%outputs
	  if (.not.output .and. .not.associated(np,fn%inputs))  np%parent => fn%inputs
	end if
	return
	end subroutine nesttype

!!	Read a single field and introduce the associated logic
!!	in a netput descriptor:

	subroutine readelem(np, f, taxagent, lastfield)
	type (netput), pointer :: np	!! Netput to which this should be assigned
	type (field), pointer  :: f	!! Field for which logic is to be introduced
	type (field), pointer :: taxagent	!! Field which defines applicable tax agent field
	type (field), pointer :: lastfield	!! Pointer to the previous field in this record (if one)
	type (taxlist), pointer :: t

	select case (f%label)

	case ('p')
	  if (associated(np%pref)) stop 'P: field appears more than once.'
	  np%pref => f

	case ('q')
	  if (associated(np%qref)) stop 'Q: field appears more than once.'
	  np%qref => f

	case ('a')
	  taxagent => f

	case ('t')
	  if (.not.associated(taxagent)) stop 'Tax field appears prior to specification of a tax agent (A:).'
	  call addtax(np,t)
	  t%t => f
	  t%a => taxagent

	case ('n')
	  if (.not.associated(taxagent)) stop 'Tax field appears prior to specification of a tax agent (A:).'
	  call addtax(np,t)
	  t%n => f
	  t%a => taxagent
	  
	case ('m')
	  if (.not.associated(taxagent)) stop 'Tax field appears prior to specification of a tax agent (A:).'
	  call lasttax(np,t)
	  if (.not.associated(lastfield))     stop 'M: field must follow N: field'
	  if (.not.associated(t%n,lastfield)) stop 'M: field must follow immediately after N: field'
	  t%m => f

	end select

	lastfield => f
	return
	end subroutine readelem

!!	Create symbol table entries for the production sectors

	subroutine sectors

	character (len=31) :: label
	type(function), pointer :: fn
	type(record), pointer :: r
	type(field), pointer :: f
	type(symbol), pointer :: s
	type(domain), pointer :: d

        fn => functionlist
        nullify(f);
        do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  if (log) write(*,'(a)') ' '//trim(f%label)//':'//trim(f%data)
	  fn => fn%next
        end do

	nullify(f);
	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  if (log) write(*,'(a)') ' label: '//trim(f%label)
	  if (f%label.eq.'$prod') then
	    if (log) write(*,'(a)') ' Adding sector: '//trim(f%identtext)
	    call findid(f%identtext, s)
	    if (associated(s)) stop ' *** Multiple $prod blocks for single sector not supported.'
	    call addid(f%identtext, 'sector', s, f%line)
	    s%dim = dimension(f%arg)
	    if (f%spantext.ne.' ') stop ' *** Span notation "#" not permited in $prod block.'
	    d => f%arg
	    do while (associated(d))
	      if (d%s%name(1:1).eq.'"' .or. d%s%name(1:1).eq."'") &
		stop ' *** Singleton sets not permitted in $prod domain'
	      d => d%next
	    end do
	  end if
	  fn => fn%next
	end do
	return
	end subroutine sectors

!!	Determine which a string is numeric:

	logical function numeric(string)
	character (len=*) :: string	!! String to be evaluated (.true. if a Fortran-readable numeric value)
	real :: testval
	integer :: stat
	if (index('.+-0123456789', string(1:1)).gt.0) then
	  read(string,*,iostat=stat) testval
	  if (stat.ne.0) stop ' ** Error in numeric field.'
	  numeric = .true.
	else
	  numeric = .false.
	end if
	return
	end function numeric

!!	Add a netput to a nest.

	subroutine addnetput(np, n, f)
	type (netput), pointer :: np	!! Netput to be added
	type (nest), pointer :: n	!! Nest in which the netput is to be introduced
	type (field), pointer :: f	!! Field in which the nest assignment is made (for condition if any)
	type (netputlist), pointer :: nl
	if (.not.associated(n%elements)) then
	  nullify(nl)
	  allocate(nl)
	  n%elements => nl
	else
	  nl => n%elements
	  do while (associated(nl%next))
	    nl => nl%next
	  end do
	  allocate(nl%next)
	  nl => nl%next
	endif
	nullify(nl%np)
	nullify(nl%next)
	nl%np => np
	if (associated(f)) nl%f => f
	return
	end subroutine addnetput

!!	Add a tax coefficient to the associated linked list for a production netput:

	subroutine addtax(np,t)
	type (netput), pointer :: np	!! Netput to which tax is applied
	type (taxlist), pointer :: t	!! Pointer to the tax field

	if (associated(np%t)) then
	  call lasttax(np, t)
	  call newtax(t%next)
	  t => t%next
	else
	  call newtax(np%t)
	  t => np%t
	end if
	return
	end subroutine addtax

!!	Set pointer to the last tax applied to a given netput:

	subroutine lasttax(np, t)
	type (netput), pointer :: np	!! Netput to be queried
	type (taxlist), pointer :: t	!! On output points to the last tax coefficient

	t => np%t
	do while (associated(t%next))
	  t => t%next
	end do
	return
	end subroutine lasttax


!!	Generate a pointer to the nest with a given name in a particular function:

	subroutine findnest(nestid, fn, n)
	character (len=*) :: nestid	!! String descriptor of the nest to be located
	type (function), pointer :: fn	!! Function in which to locate the nest
	type (nest), pointer :: n	!! Pointer to the located nest

	n => fn%inputs
	if (log) write(*,'(a)') ' Locating nest: '//trim(nestid)
	do while (associated(n))
	  if (nestid.eq.n%ident) return
	  n => n%next
	end do
	stop 'nestid not found'
	end subroutine findnest
	    
!!	At a subnest to a nest.  This requires creation of a nestlist if
!!	one does not already exist

	subroutine addsubnest(n, np)
	type (nest), pointer :: n	!! Nest to be added as a subnest
	type (nest), pointer :: np	!! Nest to which the subnest is to be introduced
	type (nestlist), pointer :: nl

	if (.not.associated(np%subnests)) then
	  nullify(nl)
	  allocate(nl)
	  np%subnests => nl
	  nullify(nl%next)
	  nl%n => n
	else
	  nl => np%subnests
	  do while (associated(nl%next))
	    nl => nl%next
	  end do
	  allocate(nl%next)
	  nl => nl%next
	  nullify(nl%next)
	  nl%n => n
	end if
	return
	end subroutine addsubnest



end module mpsge

