$title BoP model

* ----------------------------------------------------------------------------
* Small open economy allowing trade deficit or surplus to vary.
* Laurent Cretegny, CoPS, Monash University, Australia, 2004.
* ----------------------------------------------------------------------------

$if not exist bop.gdx $call gdxxrw i=bop.xls o=bop.gdx index=declare

sets
  o                 commodity and sector
  f                 primary factor;

alias (o,s);

parameters
  sam(*,*)          benchmark social accounting matrix
  supply(s)         supply of output [[sppl:s]]
  interm(o,s)       intermediate demand [[intr:o*s]]
  factor(f,s)       factor demand [[fctr:f*s]]
  demand(s)         demand for consumption [[dmnd:s]]
  govdem(s)         government demand [[gvdm:s]]
  endow(f)          endowment [[endw:f]]
  export(s)         export (at fob price) [[expr:s]]
  import(s)         import (at cif price) [[impr:s]]
  bbop              balance of payments at the benchmark (surplus) [[bbop:]]
  etrn(s)           elasticity of transformation [[etrn:s]]
  esub(s)           elasticity of substitution [[esub:s]]
  esubc             elasticity of substitution for consumption [[esbc:]]
  esubm(s)          elasticity of substitution for import [[esbm:s]]
  bte(s)            tax rate on export [[bte:s]]
  btm(s)            tax rate on import [[btm:s]]
  bpe(s)            benchmark price for export [[bpe:s]]
  bpm(s)            benchmark price for import [[bpm:s]]
  bopi              benchmark balance of payments index [[bop:]];

$gdxin bop
$load o f etrn esub esubc esubm sam
$gdxin

interm(o,s) = sam(o,s);
factor(f,s) = sam(f,s);
demand(s) = sam(s,"ra");
govdem(s) = sam(s,"gov");
endow(f) = sam("ra",f);
export(s) = sam(s,"row");
import(s) = sam("row",s);
supply(s) = sum(o, interm(o,s)) + sum(f, factor(f,s))
            - export(s) - sam(s,"trow");
bbop = sam("row","ra");
bte(s)$export(s) = - sam(s,"trow") / export(s);
btm(s)$import(s) = sam("trow",s) / import(s);
bpe(s) = 1 - bte(s);
bpm(s) = 1 + btm(s);
bopi = 1;

$if not exist bop.har $include gdxtohar

$ontext
$model:bp

$sectors:
  y(s)              ! production
  a(s)              ! aggregate supply
  u                 ! utility index
  g                 ! government expenditure

$commodities:
  pd(s)             ! domestic price of commodity
  pa(s)             ! price of commodity
  pf(f)             ! price of primary factor
  pu                ! price index for utility
  pg                ! price of a government output unit
  pfx               ! real exchange rate

$consumers:
  ra                ! representative agent income
  gov               ! government income

$auxiliary:
  bop               ! balance of payments index

$prod:y(s)          t:etrn(s)           va:esub(s)
  o:pd(s)           q:supply(s)
  o:pfx             q:export(s)         p:bpe(s)  a:gov     t:bte(s)
  i:pa(o)           q:interm(o,s)
  i:pf(f)           q:factor(f,s)       va:

$prod:a(s)          s:esubm(s)
  o:pa(s)           q:(supply(s)+bpm(s)*import(s))
  i:pd(s)           q:supply(s)
  i:pfx             q:import(s)         p:bpm(s)  a:gov     t:btm(s)

$prod:u             s:esubc
  o:pu              q:(sum(s, demand(s)))
  i:pa(s)           q:demand(s)

$prod:g
  o:pg              q:(sum(s, govdem(s)))
  i:pa(s)           q:govdem(s)

$demand:ra
  d:pu
  e:pf(f)           q:endow(f)
  e:pfx             q:-bbop             r:bop

$demand:gov
  d:pg

$constraint:bop
  u =e= 1;

$report:
  v:rsupply(s)      o:pd(s)             prod:y(s)
  v:rexport(s)      o:pfx               prod:y(s)
  v:rimport(s)      i:pfx               prod:a(s)
  v:rdemand(s)      i:pa(s)             prod:u

$offtext
$sysinclude mpsgeset bp

* Initialize levels on auxiliary variables

bop.lo = -inf;
bop.l = bopi;

* Benchmark replication

bp.iterlim = 0;
$include bp.gen
solve bp using mcp;
abort$(abs(bp.objval) gt 1e-7) "*** bop does not calibrate ! ***";
bp.iterlim = 1000;

* Fixation of a numeraire to allow comparison

pu.fx = pu.l;

* Counterfactual : free trade on export

bte(s) = 0;

$include bp.gen
solve bp using mcp;

* Results

parameters
  res_com           results by commodity in % change from basis
  res_agg           aggregate results in % change from basis; 

res_com("supply",s) = (rsupply.l(s)/supply(s) - 1) * 100;
res_com("export",s) = (rexport.l(s)/export(s) - 1) * 100;
res_com("import",s) = (rimport.l(s)/import(s) - 1) * 100;
res_com("demand",s) = (rdemand.l(s)/demand(s) - 1) * 100;
res_com("price",s) = (pa.l(s) - 1) * 100;

res_agg("u") = (u.l - 1) * 100;
res_agg("pu") = (pu.l - 1) * 100;
res_agg("pfx") = (pfx.l - 1) * 100;
res_agg("ra_inc") = (ra.l/(sum(s, demand(s))) - 1) * 100;

option decimals = 6;
display res_com, res_agg;
