$title BoP model

* ----------------------------------------------------------------------------
* Small open economy allowing trade deficit or surplus to vary.
* Laurent Cretegny, CoPS, Monash University, Australia, 2004.
* ----------------------------------------------------------------------------

$if not exist bop.gdx $call gdxxrw i=bop.xls o=bop.gdx index=declare

sets
  o                 commodity and sector
  f                 primary factor;

alias (o,s);

parameters
  sam(*,*)          benchmark social accounting matrix
  supply(s)         supply of output [[sppl:o]]
  interm(o,s)       intermediate demand [[intr:o*o]]
  factor(f,s)       factor demand [[fctr:f*o]]
  demand(o)         demand for consumption [[dmnd:o]]
  govdem(o)         government demand [[gvdm:o]]
  endow(f)          endowment [[endw:f]]
  export(o)         export (at fob price) [[expt:o]]
  import(o)         import (at cif price) [[impt:o]]
  bbop              balance of payments at the benchmark (surplus)
  etrn(s)           elasticity of transformation [[etrn:o]]
  esub(s)           elasticity of substitution [[esub:o]]
  esubc             elasticity of substitution for consumption [[esbc:]]
  esubm(o)          elasticity of substitution for import [[esbm:o]]
  bte(o)            tax rate on export [[te:o]]
  btm(o)            tax rate on import [[tm:o]]
  bpe(o)            benchmark price for export
  bpm(o)            benchmark price for import;

$gdxin bop
$load o f etrn esub esubc esubm sam
$gdxin

interm(o,s) = sam(o,s);
factor(f,s) = sam(f,s);
demand(o) = sam(o,"ra");
govdem(o) = sam(o,"gov");
endow(f) = sam("ra",f);
export(o) = sam(o,"row");
import(o) = sam("row",o);
supply(s) = sum(o, interm(o,s)) + sum(f, factor(f,s))
            - export(s) - sam(s,"trow");
bbop = sam("row","ra");
bte(o)$export(o) = - sam(o,"trow") / export(o);
btm(o)$import(o) = sam("trow",o) / import(o);
bpe(o) = 1 - bte(o);
bpm(o) = 1 + btm(o);

$if not exist bop.har $include gdxtohar

$ontext
$model:bp

$sectors:
  y(s)              ! production
  va(s)             ! value added
  a(s)              ! aggregate supply
  u                 ! utility index
  g                 ! government expenditure

$commodities:
  pd(o)             ! domestic price of commodity
  pa(o)             ! price of commodity
  pva(s)            ! price index for value added
  pf(f)             ! price of primary factor
  pu                ! price index for utility
  pg                ! price of a government output unit
  pfx               ! real exchange rate

$consumers:
  ra                ! representative agent income
  gov               ! government income

$auxiliary:
  bop               ! balance of payments index

$prod:y(s)          t:etrn(s)
  o:pd(s)           q:supply(s)
  o:pfx             q:export(s)         p:bpe(s)  a:gov     t:bte(s)
  i:pa(o)           q:interm(o,s)
  i:pva(s)          q:(sum(f, factor(f,s)))

$prod:va(s)         s:esub(s)
  o:pva(s)          q:(sum(f, factor(f,s)))
  i:pf(f)           q:factor(f,s)

$prod:a(o)          s:esubm(o)
  o:pa(o)           q:(supply(o)+bpm(o)*import(o))
  i:pd(o)           q:supply(o)
  i:pfx             q:import(o)         p:bpm(o)  a:gov     t:btm(o)

$prod:u             s:esubc
  o:pu              q:(sum(o, demand(o)))
  i:pa(o)           q:demand(o)

$prod:g
  o:pg              q:(sum(o, govdem(o)))
  i:pa(o)           q:govdem(o)

$demand:ra
  d:pu
  e:pf(f)           q:endow(f)
  e:pfx             q:-bbop             r:bop

$demand:gov
  d:pg

$constraint:bop
  u =e= 1;

$report:
  v:rsupply(s)      o:pd(s)             prod:y(s)
  v:rexport(o)      o:pfx               prod:y(o)
  v:rimport(o)      i:pfx               prod:a(o)
  v:rdemand(o)      i:pa(o)             prod:u

$offtext
$sysinclude mpsgeset bp

* Initialize levels on auxiliary variables

bop.lo = -inf;
bop.l = 1;

* Benchmark replication

bp.iterlim = 0;
$include bp.gen
solve bp using mcp;
abort$(abs(bp.objval) gt 1e-7) "*** bop does not calibrate ! ***";
bp.iterlim = 1000;

* Fixation of a numeraire to allow comparison

pu.fx = pu.l;

* Counterfactual : free trade on export

bte(o) = 0;

$include bp.gen
solve bp using mcp;

* Results

parameters
  res_com           results by commodity in % change from basis
  res_agg           aggregate results in % change from basis; 

res_com("supply",s) = (rsupply.l(s)/supply(s) - 1) * 100;
res_com("export",o) = (rexport.l(o)/export(o) - 1) * 100;
res_com("import",o) = (rimport.l(o)/import(o) - 1) * 100;
res_com("demand",o) = (rdemand.l(o)/demand(o) - 1) * 100;
res_com("price",o) = (pa.l(o) - 1) * 100;

res_agg("u") = (u.l - 1) * 100;
res_agg("pu") = (pu.l - 1) * 100;
res_agg("pfx") = (pfx.l - 1) * 100;
res_agg("ra_inc") = (ra.l/(sum(o, demand(o))) - 1) * 100;

option decimals = 6;
display res_com, res_agg;
