! KP question for Tom
! It seems that for an "s:" field, %ident is not set up - see results from
!  KP's routine GetProdTopVariableElast
! fixed we think.

! KP question for Tom/Laurent
!  What are the rules for figuring out the market clearing equations?

! KP question for Tom/Laurent
!  In Laurent's basic.mps, has p(s) and p(o) and alias (s,o). Can we insist that
!   no alias is used in the MPSGE description?
! When have p(s) and p(o), GEMPACK needs to know that o is subset of s and/or
!   vice versa.
! Must we exclude elements occurring in the .MPS file?
! Answer. Tom will read declarations and all variables (not quantities - parameters)
!  or sets in : fields must have been declared.

! KP question for Tom/Laurent
! How do I put comments in the .MPS file?
! Answer. ! before trailing comment or * in first column

! KP question.
! What are the implied balancing conditions? Should we test them?
! Answer. The coeffs on the 2 sides of each market clearing equation must be equal.
! That is, when set variables equal to 1, get equality.
! Perhaps calc differences and tell user to check they are small.

! Restriction we may need to impose
! $prod may have two outputs - either two o: fields or 
!  one o: ranging over different set from $prod: field.
! Do we ban this?
! We have to figure out the other restrictions we need to impose. 

!!     Routines for constructing and manipulating general equilibrium
!!      models implemented with MPSGE
!!      @version 1.0
!!      @author Thomas Rutherford

module mpsge

        implicit none

!       Records of source data:

!!      Defines a line of MPSGE source code.

        type line
          character (len=2048) :: text		!! Input records are limited to 2048 characters
          integer :: lineno			!! Line number in the input file
	  type (record), pointer :: record	!! Record to which this line belongs
          type (line), pointer :: next		!! Pointer to the next line
        end type line

!!      Defines a list of input lines as they have been read from the input file.

        type linelist
          type (line), pointer :: line
          type (linelist), pointer :: next
        end type linelist

!!      Elements of the symbol table include various MPSGE variables (commodity
!!      prices, sectoral activity level, consumer incomes and auxiliary variables).
!!      The symbol table also includes sets and parameters which are encountered
!!      in the input file.  This list is maintained to provide some context checking
!!      on input.

        type symbol
          character (len=31) :: name		!! Symbol identifier
          character (len=9) :: type		!! One of set, parameter, sector, commodity, consumer, auxiliary
          type (linelist), pointer :: lines     !! Points to lines in which this symbol appears
          type (symbol), pointer :: next        !! Points to next symbol in the global list
          integer :: dim			!! Dimension of the symbol (undefined=-1,scalar=0, vector=1, etc.)
! 28/01/04 KP thinks that DeclaredGP is probably no longer needed
	  logical :: DeclaredGP                 !! Whether this has been declared in the GEMPACK TAB file
        end type symbol

!!      A domain is simply a list of sets.  This list may include quoted singletons as well as
!!      references to GAMS sets.  Each element of the set is represented by a pointer to the
!!      set in the symbol table

        type domain
          type (symbol), pointer :: s           !! Pointer to this set in the symbol table
          type (domain), pointer :: next        !! Pointer to the next set in this domain (unassociated if no more dimensions)

          integer :: dim			!! Dimension of this domain
        end type domain

!!      An MPSGE input file consists of a series of records.  Each record is composed of
!!      one or more fields.  An input field consist of a label and some field text.

!!      A field in an MPSGE model consists of two primary components: the label and
!!      the field data.  The field data in turn consists of the identifier,
!!      the identifier argument, the spanning sets and the condition clause.

        type field

!! Verbatim text of leading label

          character (len=255)   :: label

!! Verbatim text of following field 

          character (len=255)   :: data

!! Identifier text (data excluding domain and condition)

          character (len=255)   :: identtext

!! Substring text of the identitier argument domain if present

          character (len=255)   :: argtext

!! Substring text of the spanning domain if present (# notation)

          character (len=255)   :: spantext

!! Substring text of the conditional clause if present

          character (len=255)   :: condtn

!! Pointer to field idenfier if the field is readable. 
!! Fields beginning with "(" are skipped.

          type (symbol), pointer :: ident       

!! Pointer to argument domain if it is readable

          type (domain),pointer :: arg

!! Points to span domain if it is present

          type (domain),pointer :: span         

!! Points to source line

          type (line), pointer :: line          

!! Pointer to next field in the data file.
!! The final field in a datafile is unassociated.

          type (field), pointer :: next
                                                
        end type field

!!      This type describes a _record_ in an MPSGE function, possibly spanning multiple lines
!!      if there are continuations (lines "+" in the first column)

        type record
          type (field), pointer  :: firstfield  !! Points to the first field in this record
          type (line), pointer   :: line        !! Points to the input line where this record begins.
          integer                :: nfield      !! Counts the number of fields in this record
          type (record), pointer :: next        
        end type record

!!      A tax coefficient on a production input or output consists of:
!!      - A tax revenue agent (identified in an A: field)
!!      - An ad-valorem tax rate defined by one of the following:
!!        - A T: field representing a fixed and exogenous tax rate
!!        - A N:/M: field pair representing an endogenous tax.
!!      The tax coefficient consists only of pointers to fields in the input file.  Note that 
!!      the A: field may be on a record which appears prior to the tax rate field.
!!      With endogenous taxes, the N: field defines an auxiliary variable which controls the tax
!!      and the M: field defines a multiplier on the tax rate.

        type taxlist

!!      Pointer to A: field which identifes agent who collects this tax revenue:

              type (field), pointer :: a

!!      Pointer to T: field which defines an exogenous ad-valorem tax rate:

              type (field), pointer :: t

!!      Pointer to N: field indicating the auxiliary variable which determines the
!!      ad-valorem tax rate.  For a given tax entry, either T: or N: will be
!!      used, but not both:

              type (field), pointer :: n

!!      Pointer to an M: field defining the multiplier which is to be applied on 
!!      the previously listed endogenous tax.  An M: field must follow immediately
!!      after an N: field.

              type (field), pointer :: m

!!      Next tax applied on the same production input or output.  This is unassociated
!!      if there are no more applicable taxes:

              type (taxlist), pointer :: next
        end type taxlist

!!      This type defines an input or output in a production block.  The domain associated with
!!      a netput consists of the union of sets in the function declaration block and the 
!!      commodity price.  For example, if we have
!!>
!!      $prod:y(i,t)
!!              o:p(j,t)        q:y0(i)
!!              ...
!!<
!!      then the _domain_ of the netput, _d_, is (_i,t,j_), and the _subdomain_ of the netput, _sd_, is (_j_).
!!      

        type netput
                type (field), pointer    :: f           !! Pointer to the I: or O: field
                type (field), pointer    :: qref        !! Pointer to the associated Q: field in the same record
                type (field), pointer    :: pref        !! Pointer to the associated P: field in the same record
                type (domain), pointer   :: d           !! Pointer to the domain of this netput
                type (domain), pointer   :: sd          !! Pointer to the subdomain of the netput
                type (taxlist), pointer  :: t           !! Pointer to the taxes which apply to this netput
		character (len=1024) :: nestcount
		integer :: nestfields
		logical :: output			!! Logical flag for an output coefficient
                integer :: index                        !! Integer index associated with this netput
        end type netput

        type endowment
                type (field), pointer    :: f           !! Pointer to the E: field
                type (field), pointer    :: q           !! Pointer to the Q: field
		type (field), pointer    :: r		!! Pointer to the R: field
                type (domain), pointer   :: d           !! Pointer to the domain of the endowment
                type (domain), pointer   :: sd          !! Pointer to the subdomain of the endowment
		integer :: index
        end type endowment

!!	List of endowment records

	type endowmentlist
		type (endowment), pointer     :: e	
		type (endowmentlist), pointer :: next
	end type endowmentlist


!!      List of netputs references from the associated nest.

        type netputlist
                type (netput), pointer :: np            !! Netput at this location in the list
                type (field),  pointer :: f             !! Nest assignment field
                type (netputlist), pointer :: next      !! Pointer to next element in list (nul if last input)
        end type netputlist

!!      Nest type characterizes a CES nest in a cost function.

        type nest
                character (len=255) :: ident		!! Identifier text for this nest 
                character (len=255) :: parentid		!! Identifier text for the parent nest
                character (len=255) :: value            !! Text field of the elasticity value
                logical :: output                       !! Logical flag identifies type of nest (output or input)
                type (field), pointer :: f              !! Pointer to the input field which defines the nest
                type (nest), pointer :: parent          !! Pointer to parent nest
                type (nestlist), pointer :: nlist       !! Pointer to parent nestlist entry
                type (netputlist), pointer :: elements  !! Pointer to a list of constituent inputs or outputs
                type (nestlist), pointer :: subnests    !! Pointer to a list of constituent nests
                type (nest), pointer :: next            !! Pointer to next nest in current function
                integer :: index                        !! Integer index associated with this nest
		integer :: level			!! Level in the nesting structure (top level = 0)
		logical :: assigned			!! Flag that all the logic for this nest has been written
	end type nest

!!      Linked list of nests used to represent the nested CES substitution structure:

        type nestlist
                type (nest), pointer :: n               !! Pointer to a nest in this list
                type (nestlist), pointer :: next        !! Pointer to next element of the list
        end type nestlist

!! An MPSGE production or demand function

        type function           
          type (record), pointer :: firstrecord !! Pointer to first input record in this function
          integer               :: nrecord      !! Count of number of records in this function
          integer               :: nfield       !! Count of number of fields in this function
          type (nest), pointer  :: inputs       !! Points to the top level input nest    
          type (nest), pointer  :: outputs      !! Points to the top level output nest    
	  type (field), pointer :: demand		!! Points to the demand field
	  type (endowmentlist), pointer :: endowments	!! Points to the endowments declared in this $demand block
          type (function), pointer :: next      !! Pointer in linked list of all functions
        end type function

        type (line), pointer     :: inputfile   !! Pointer to first line in the linked list of input file lines
        type (function), pointer :: functionlist !! Pointer to the first function in linked list
        type (symbol), pointer   :: idlist      !! Pointer to the first symbol
        type (symbol), pointer   :: idlast      !! Pointer to the last symbol

        character (len=32) :: modelname

        integer :: nfun=0               !! Number of functions in the MPSGE file
        integer, parameter :: iomps=10  !! Unit number of the MPSGE file
        integer, parameter :: iogms=11  !! Unit number of the resulting GAMS file

	logical :: splitlog=.false.	!! Flag for log output of field parsing operation
	logical :: log=.true.		!! Flag for general log output
        logical :: readvars         !! Flag for current input ot declaration records

! Declarations/Types added by KP

	type GPCoeffVarReal    ! GEMPACK Coefficient or Variable or Real
	  character :: type  ! 'c' for coefficient or 'v' for variable or 'r' for real
	  character (len=15) :: name    ! Coeff/Var name or Real as string
	  type (domain), pointer :: arg      ! Sets over which arguments range
	end type GPCoeffVarReal

        integer, parameter :: iotabgp=11   !! Unit number of the output GEMPACK TAB file
        integer, parameter :: iocmfgp=12   !! Unit number of the output GEMPACK Command file
	integer :: NHeaderGP=0          !! Number of headers data is read from in TAB file

	integer LengthGPDeclarationName
	parameter (LengthGPDeclarationName = 20)
	type GPDeclaration   ! List of various types of GEMPACK declarations
	  character (len=2) :: type  !! 'cs' for Coeff, 'vc' for Variable, 'eq' for equation
	  character (len=LengthGPDeclarationName) :: name  !! Coeff/Var/EQuation name
	  type (GPDeclaration), pointer :: next !! Pointer to the next declaration in this list
	end type GPDeclaration

	type (GPDeclaration), pointer :: GPDeclarationList !! Pointer to first in list 
	type (GPDeclaration), pointer :: GPDeclarationLast !! Pointer to last in list

contains

!!      Master routine to read an MPSGE input file:
!!        - Open the file
!!        - Read one line at a time into a linked list for listing.
!!        - Locate functions and set up pointers to the constituent records.
!!        - Parse records into fields.
!!
!!	None of the logical structure of the underlying economic model is
!!	formulated at this stage.

        subroutine readmps
        integer :: lineno
	type (line), pointer :: L
        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f, ff
        character (len=2048) :: input=' '
        type(function), pointer :: fn2 ! KP

        nullify(inputfile)
	nullify(functionlist)
	nullify(f)
	nullify(L)

	write(iogms,'(a)') '$ontext'

        lineno = 0
        do

!       Read a line until we reach one which is not a comment:
!	L is nullified at end of file.

	  call readline(L, input, lineno)
	  if (.not.associated(L)) exit

!       Read a new function?

          if (input(1:1).eq.'$') then
            nfun = nfun + 1
            if (associated(functionlist)) then
	      write(*,*) 'Newfunction 1. NFUN=', nfun   ! KP
	      call newfunction(fn%next); fn => fn%next
	      fn2 => functionlist   ! KP
	      write(*,*) 'FN2 is:', fn2%firstrecord%firstfield%label   ! KP
!NO	      fn2 => fn2%next   ! KP
!NO	      write(*,*) 'FN2%NEXT is:', fn2%firstrecord%firstfield%label   ! KP
            else
	      write(*,*) 'Newfunction 2. NFUN=', nfun   ! KP
	      call newfunction(functionlist); fn => functionlist
            endif

!	Set a flag here if we are reading variable declarations to avoid
!	pulling the plug if we can't find a colon:

	    if (eqv(input,'$sectors:') .or. eqv(input,'$commodities:') .or. &
		eqv(input,'$consumers:') .or. eqv(input,'$auxiliary:')) then
	      readvars = .true.
	    else
	      readvars = .false.
	    end if
          else
            if (.not.associated(functionlist)) &
              stop 'Error == MPSGE input must begin with a function.'
          end if

!       Read a new record?

          if (input(1:1).ne.'+') then
            if (associated(fn%firstrecord)) then
	      call newrecord(r%next);  r => r%next; 
            else
	      call newrecord(r); fn%firstrecord => r; 
            end if
            r%line => L
            fn%nrecord = fn%nrecord + 1
          end if
	  nullify(f)
	  L%record => r

!	Keep pulling fields off the input record until there are no more:

          do while (input.ne.' ')

	    call newfield(ff)

!	    if (associated(lastfield)) lastfield%next => ff

!	This logic only applies to the first field extracted from the
!	line:

            if (input(1:1).ne.'+') r%firstfield => ff
	    if (associated(f)) ff%next => f

            f => ff
            f%line => L
            r%nfield = r%nfield + 1
	    fn%nfield = fn%nfield + 1

	    call readfield(input,f)
          end do

        end do

        write(iogms,'(a)') '$offtext'
	write(*,*) ' NFUN= ', nfun    ! KP
	fn2 => functionlist   ! KP
	write(*,*) 'FN2 is:', fn2%firstrecord%firstfield%label   ! KP
	fn2 => fn2%next   ! KP
	write(*,*) 'FN2%NEXT is:', fn2%firstrecord%firstfield%label   ! KP
	fn2 => fn2%next   ! KP
	write(*,*) 'FN2%NEXT%NEXT is:', fn2%firstrecord%firstfield%label   ! KP

	return
        end subroutine readmps

!!	Read to the next non-blank line from the input file, incrementing
!!	the line count:

	subroutine readline(L, input,lineno)
	type (line), pointer :: L
        character (len=*) :: input
	integer :: lineno
	integer :: ic

	input = ' '
	do while (input.eq.' ') 
          read(iomps,'(a)',end=100) input
	  lineno = lineno + 1
          write(iogms,'(a,i0,a,t12,a)') '[',lineno,']',trim(input)
            
	  if (associated(inputfile)) then
	    call newline(L%next)
	    L => L%next
	  else
	    call newline(inputfile)
	    L => inputfile
	  endif
	  L%text = input
	  L%lineno = lineno
            
	  ic = index(input,'!')
	  if (ic.gt.0) input(ic:) = ' '
	  call tab2blank(input)       ! Convert tabs to blanks:
	  if (input(1:1).eq.'*') input = ' '
	end do
	return
100	nullify(L)
	return
	end subroutine readline

!!	Read the last field from a line of input, using the 
!!	colon to locate the label and the data components:	

	subroutine readfield(input,f)
	type (field), pointer :: f
        character (len=*) :: input
	integer :: ic

	ic = index(input,':',back=.true.)
	if ((.not.readvars) .and. (ic.eq.0)) stop 'No colon?'
	f%data = adjustl(input(ic+1:))
	input(ic:) = ' '
	ic = index(input(1:ic-1),' ',back=.true.)
	f%label = input(ic+1:)
	call lower(f%label)
	input(ic+1:) = ' '
	if (input.eq. '+') input = ' '
	return
	end subroutine readfield

!!      Fold a character string to lower case.

        subroutine lower(word)
        IMPLICIT NONE
        character (len=*) :: word  !! Character to be placed in lower case.
        integer :: i
        do 100 i=1,len(word)
          if (word (i:i).eq.'A') word (i:i) = 'a'
          if (word (i:i).eq.'B') word (i:i) = 'b'
          if (word (i:i).eq.'C') word (i:i) = 'c'
          if (word (i:i).eq.'D') word (i:i) = 'd'
          if (word (i:i).eq.'E') word (i:i) = 'e'
          if (word (i:i).eq.'F') word (i:i) = 'f'
          if (word (i:i).eq.'G') word (i:i) = 'g'
          if (word (i:i).eq.'H') word (i:i) = 'h'
          if (word (i:i).eq.'I') word (i:i) = 'i'
          if (word (i:i).eq.'J') word (i:i) = 'j'
          if (word (i:i).eq.'K') word (i:i) = 'k'
          if (word (i:i).eq.'L') word (i:i) = 'l'
          if (word (i:i).eq.'M') word (i:i) = 'm'
          if (word (i:i).eq.'N') word (i:i) = 'n'
          if (word (i:i).eq.'O') word (i:i) = 'o'
          if (word (i:i).eq.'P') word (i:i) = 'p'
          if (word (i:i).eq.'Q') word (i:i) = 'q'
          if (word (i:i).eq.'R') word (i:i) = 'r'
          if (word (i:i).eq.'S') word (i:i) = 's'
          if (word (i:i).eq.'T') word (i:i) = 't'
          if (word (i:i).eq.'U') word (i:i) = 'u'
          if (word (i:i).eq.'V') word (i:i) = 'v'
          if (word (i:i).eq.'W') word (i:i) = 'w'
          if (word (i:i).eq.'X') word (i:i) = 'x'
          if (word (i:i).eq.'Y') word (i:i) = 'y'
          if (word (i:i).eq.'Z') word (i:i) = 'z'
          if (word (i:i).eq.char(9)) word(i:i) = ' '
100     continue
        return
        end subroutine lower

!!      Fold a character string to upper case.

        subroutine upper(word)
        IMPLICIT NONE
        character (len=*) :: word  !! Character to be placed in upper case.
        integer :: i, iacharpos, newpos
	character :: ch1
        do 100 i=1,len(word)
	  ch1 = word(i:i)
	  iacharpos=IACHAR(ch1)
! "a" is ASCII character 97 and "z" is ASCII character 122
! "A" is ASCII character 95 and "z" is ASCII character 90
	  if(iacharpos.ge.97.and.iacharpos.le.122) then
	    newpos = iacharpos - 32
	    word(i:i) = ACHAR(newpos)
	  end if
          if (word (i:i).eq.char(9)) word(i:i) = ' '
100     continue
        return
        end subroutine upper

!!      Convert tabs on an input line to spaces:

        subroutine tab2blank(line)
        character (len=*) :: line  !! Line to scan for tabs
        character :: tab=char(9)
        integer :: k
        do k=1,len_trim(line)
          if (line(k:k).eq.tab) line(k:k) = ' '
        end do
        return
        end subroutine tab2blank

!!      Evaluate the dimension of a domain by counting
!!      the number of set identifiers in the linked list
!!      defining the domain:

        integer function dimension(d)
        type (domain), pointer :: d     !! Domain to be evaluated
        type (domain), pointer :: dd    
        dimension = 0
        if (.not.associated(d)) return
        dd => d
        do while (associated(dd))
          dimension = dimension + 1
          dd => dd%next
        end do
        end function dimension

!!      Generate an echoprint of the file.  This is currently very preliminary,
!!      and will later be recoded to produce a more carefully formatted report.

        subroutine echoprint

        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f
	integer :: i

        fn => functionlist
        nullify(f);
        do while (associated(fn))
          write(*,'(a)') ' ===== '
	  r => fn%firstrecord
	  do while (associated(r))
	    f => r%firstfield
	    do while (associated(f))
	      write(*,'(a)') ' '//trim(f%label)//':'//trim(f%data)
	      write(*,'(a)') ' Identifier text: '//trim(f%identtext) ! KP
	      write(*,'(a)') ' Argument text: '//trim(f%argtext) ! KP
	      f => f%next
	    end do
	    r => r%next
	  end do
	  fn => fn%next
        end do

	return
	end subroutine echoprint

!!      Parse all fields in the file into constituent elements:
!!      - Field identifier text
!!      - Identifier argument text
!!      - Spanning domain text
!!      - Condition text.

        subroutine parse
        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f
        type(symbol), pointer :: s

        fn => functionlist
        nullify(f);
        do while (associated(fn))
	  r => fn%firstrecord
	  do while (associated(r))
	    f => r%firstfield
	    do while (associated(f))

!	Sort out the data:

	      call split(f%data, f%identtext, f%argtext, f%spantext, f%condtn)

!       Read domain for each field:

	      if (f%argtext .ne.' ') call readdomain(f%argtext,  f%arg, f%line)
	      if (f%spantext.ne.' ') call readdomain(f%spantext, f%span, f%line)
! Addition 29/01/04 by KP to set f%ident in all cases
	      if (f%identtext.ne.' ') then
		call findid(f%identtext, s)
		f%ident => s
	      end if
	      f => f%next
	    end do
	    r => r%next
	  end do
	  fn => fn%next
	end do
	return
        end subroutine parse

!! Return pointer to given identifier

        subroutine findid(name, id)             
        character (len=*), intent(in) :: name   !! Identifier for the symbol to be located
        type (symbol), pointer :: id            !! Pointer into the master list for the located symbol.
        type (symbol), pointer :: s

        nullify(id)
        s => idlist
        do while (associated(s))
          if (s%name==name) then
            id => s
            return
          end if
          s => s%next
        end do
        return 
        end subroutine findid

!! Introduce a new identifier in the master list.

        subroutine addid(name, type, s, L)      

!!      Name of the identifier to be added to the master list

        character (len=*), intent(in) :: name

!!      Description of the type of item to be introduced:
!!      - _set_: a GAMS set identifier
!!      - _parameter_ : a GAMS parameter identifier
!!      - _sector_ : an MPSGE sector (acivity variable)
!!      - _commodity_ : an MPSGE commodity (price variable)
!!      - _consumer_ : an MPSGE consumer (income variable)
!!      - _auxiliary_ : an MPSGE auxiliary variable

        character (len=*), intent(in) :: type
        type (symbol), pointer :: s
        type (line), pointer :: L

        write(*,'(a)') 'Adding symbol: '//trim(name)//' '//trim(type)

        if (.not.associated(idlist)) then
	  call newsymbol(idlist)
          s => idlist
        else
	  call newsymbol(idlast%next)
          s => idlast%next
        end if
        s%name = name
        s%type = type
        allocate(s%lines);
        s%lines%line => L
        nullify(s%lines%next)
        nullify(s%next)
        idlast => s
        return
        end subroutine addid

!	Create a copy of a domain.

        subroutine domaincopy(d1,d2)            
        type (domain), pointer:: d1	!! Domain to be copied
        type (domain), pointer :: d2	!! New domain to be created as a copy of d1
        type (domain), pointer :: d,dd
        if (.not.associated(d1)) return

	call newdomain(d2)
        d  => d1; dd => d2
        do while (associated(d))
          dd%s => d%s
          if (associated(d%next)) then
	    call newdomain(dd%next); dd => dd%next
          endif
          d => d%next
        end do
        return
        end subroutine domaincopy

!!	Add an entry in the linked list of line numbers which refer to a given symbol.

        subroutine addlineref(L,s)

        type (symbol), pointer  :: s  !! Symbol which is referenced
        type (line), pointer :: L	!! Line in which the symbol is referenced
        type (linelist), pointer :: LL

        LL => s%lines
        do while (associated(LL))
          if (L%lineno.eq.LL%line%lineno) return
          if (.not. associated(LL%next)) exit
          LL => LL%next
        end do
        allocate(LL%next)
        LL => LL%next
        LL%line => L
        nullify(LL%next)
        return
        end subroutine addlineref

!! Add a set element to the domain

        subroutine domainadd(name,dm, L)     
        character (len=*), intent (in) :: name	!! Identier of set to be added to the domain
        type (domain), pointer :: dm  !! Domain to which this set is to be introduced
        type (line), pointer :: L  !! Line in which the set reference was made.

        type (symbol), pointer  :: s
        type (domain), pointer :: d

!       Find what set we are introducing

        call findid(name, s)
        if (associated(s)) then
          if (s%type.ne.'set') stop 'Error -- not a SET'
          call addlineref(L, s)
        else
          call addid(name,'set',s,L)
        end if

        if (.not. associated(dm)) then
	  call newdomain(dm); dm%s => s; 
          return
        endif
        
!       Add this set at the end of the domain list:

        d => dm
        do 
          if (.not.associated(d%next)) exit
          d => d%next
        end do
	call newdomain(d%next)
	d=>d%next; d%s => s; 

        return
        end subroutine domainadd


!!  Logical function which reports whether a set with a given name
!!  is in a domain:

        logical function indomain(s,dm)
        type (symbol), pointer	:: s	!! Set to be located
        type (domain), pointer	:: dm	!! Domain in which the set is to be located.

        type (domain), pointer :: d

        if (.not. associated(dm)) then
          indomain =.false.
          return
        endif

        d => dm
        do while (associated(d))
          if (associated(d%s,s)) then
            indomain = .true.
            return
          endif
          d => d%next
        end do
        indomain = .false.
        return
        end function indomain

!!   Tets whether one domain (_d1_) is a subst of another (_d2_).

        logical function subset(d1,d2) 
        type (domain), pointer :: d1	!! Subset domain
	type (domain), pointer :: d2	!! Superset domain 
	type (domain), pointer :: d, dd
        d  => d1; 
        do while (associated(d))
          dd => d2;
          subset = .false.
          do while (associated(dd))
            if (d%s%name.eq.dd%s%name) then
              subset = .true.
              exit
            end if
            dd => dd%next
          end do
          if (.not.subset) return
          d => d%next
        end do
        return
        end function subset

!!      Split up an input field into the constituent elements
!!      - Field identifier text
!!      - Identifier argument text
!!      - Spanning domain text
!!      - Condition text.
!!
!!      For example:
!!>
!!      i:p(i,t)#(s)$beta       
!!<
!!      parses into:
!!>
!!      Field identifier: p
!!      Argument text: i,t
!!      Spanning domain text: s
!!      Condition text: beta
!!<
        subroutine split(field,ident,arg,span,condtn)
        character (len=*), intent(in)  :: field
        character (len=*), intent(out) :: ident,arg,span,condtn
        integer :: ic, np

        ident = ' '
        arg = ' '
        span = ' '
        condtn = ' '

!       Split field text into text fields for the identifier, the argument domain,
!       the span domain and conditional clause:

        np = 0
        ident = field
        do ic=1,len_trim(field)
          if (field(ic:ic).eq.'(') np = np + 1
          if (field(ic:ic).eq.')') np = np - 1
          if (ic.gt.1 .and. field(ic:ic).eq.'$' .and. np.eq.0) then
            condtn = field(ic+1:)
            ident = field(1:ic-1)
            exit
          end if
        end do
        if (np.ne.0) stop ' *** Parethesis mismatch.'

        ic = index(ident,'#')
        if (ic.gt.0) then
          span = adjustl(ident(ic+1:))
          ident(ic:) = ' '
        end if

        ic = index(ident,'(')
        if (ic.gt.1) then
          arg = ident(ic:)
          ident(ic:) = ' '
        end if

	if (splitlog) then
	  write(*,'(a)') ' --------------'
	  write(*,'(a)') ' field  = '//trim(field)
	  write(*,'(a)') ' ident  = '//trim(ident)
	  write(*,'(a)') ' arg    = '//trim(arg)
	  write(*,'(a)') ' span   = '//trim(span)
	  write(*,'(a)') ' condtn = '//trim(condtn)
	end if

        return
        end subroutine split

!!	Read a domain and parse into a series of set elements.  

        subroutine readdomain(dtext,d,L)
        character (len=*), intent(in) :: dtext	!! Domain to be parsed
        type (domain), pointer :: d  !! Pointer to the domain list
        type (line), pointer :: L  !! Pointer to source line
        character (len=255) :: textfield
        character (len=31) :: setid
        integer :: ic

        textfield = dtext

!       Check for leading and closing parentheses (required):

        if (len_trim(textfield).le.2) then
          stop ' *** READDOMAIN: invalid domain text.'
        endif

        if (textfield(1:1).ne.'(') then
          stop ' *** READDOMAIN: opening parenthes in domain text is missing.'
        endif

        if (textfield(len_trim(textfield):len_trim(textfield)).ne.')') then
          stop ' *** READDOMAIN: closing parenthesis in domain text is missing.'
        endif

        textfield = adjustl(textfield(2:len_trim(textfield)-1))

!       Check that there are no extra paretheses:

        if (index(textfield,'(') + index(textfield,')')  + &
            index(textfield,';') + index(textfield,':').gt.0) then
          stop ' *** READDOMAIN: embedded parenthesese or semicolons in domain text.'
        endif

        call lower(textfield)
        do while (textfield.ne.' ')
          ic = index(textfield,',')
          if (ic.eq.0) then
            setid = adjustl(textfield)
            textfield = ' '
          else
            if (ic.eq.1) stop ' *** READDOMAIN: empty set specification.'
            setid = adjustl(textfield(1:ic-1))
            textfield = textfield(ic+1:)
          end if
          call domainadd(setid, d, L)
        end do

        return
        end subroutine readdomain

!!	Allocate a new line and assign default values

	subroutine newline(L)
	type (line), pointer :: L
	nullify(L)
	allocate(L)
	nullify(L%next)
	L%lineno = 0
	L%text = ' '
	return
	end subroutine newline

!!	Allocate a new symbol and assign default values
!!	for all the symbol characteristics

	subroutine newsymbol(s)
	type (symbol), pointer :: s
	nullify(s)
	allocate(s)
	s%name = ' '
	s%type = 'undefined'
	nullify(s%lines)
	nullify(s%next)
	s%dim = -1
	s%DeclaredGP = .false.
	return
	end subroutine newsymbol

!!	Allocate a new function pointer with default values

	subroutine newfunction(fn)
	type (function), pointer :: fn
	nullify(fn)
	allocate(fn)
	fn%nfield = 0
	fn%nrecord = 0
	nullify(fn%next)
	nullify(fn%inputs)
	nullify(fn%outputs)
        nullify(fn%endowments)
        nullify(fn%demand)
	nullify(fn%firstrecord)
	return
	end subroutine newfunction


!!	Allocate a new domain pointer with default values

	subroutine newdomain(d)
	type (domain), pointer :: d
	nullify(d)
	allocate(d)
	nullify(d%next)
	nullify(d%s)
	d%dim = -1
	return
	end subroutine newdomain

!!	Allocate a new field with default values

	subroutine newfield(f)
	type (field), pointer :: f
	nullify(f)
	allocate(f)
	f%label = ' '
	f%data = ' '
	f%identtext = ' '
	f%argtext = ' '
	f%spantext = ' '
	f%condtn = ' '
	nullify(f%ident)
	nullify(f%arg)
	nullify(f%span)
	nullify(f%line)
	nullify(f%next)
	return
	end subroutine newfield

!!	Allocate a new record with default values

	subroutine newrecord(r)
	type (record), pointer :: r
	nullify(r)
	allocate(r)
	nullify(r%firstfield)
	nullify(r%line)
	r%nfield = 0
	nullify(r%next)
	return
	end subroutine newrecord


!!	Allocate a new tax record with default values

	subroutine newtax(t)
	type (taxlist), pointer :: t
	nullify(t)
	allocate(t)
	nullify(t%a)
	nullify(t%t)
	nullify(t%n)
	nullify(t%m)
	nullify(t%next)
	return
	end subroutine newtax

!!	Allocate a new netput with default values

	subroutine newnetput(np)
	type (netput), pointer :: np
	nullify(np)
	allocate(np)
	nullify(np%f)
	nullify(np%qref)
	nullify(np%pref)
	nullify(np%d)
	nullify(np%sd)
	nullify(np%t)
	np%nestcount = ' '
	np%nestfields = 0
	np%index = 0
	return
	end subroutine newnetput

!!	Allocate a new nest with default values

	subroutine newnest(n)
	type (nest), pointer :: n
	nullify(n)
	allocate(n)
	n%ident = ' '
	n%parentid = ' '
	n%value = ' '
	n%output = .false.
	nullify(n%f)
	nullify(n%next)
	nullify(n%parent)
        nullify(n%nlist)
	nullify(n%elements)
	nullify(n%subnests)
	n%index = 0
	n%level = 0
        n%assigned = .false.
	return
	end subroutine newnest

!!	Add d1 to union, excluding quoted singletons

	subroutine union(d1,u)
	type (domain), pointer :: d1, u, d, dd

	if (.not.associated(d1)) stop 'd1 is not associated'
	if (.not.associated(u)) nullify(u)

	d => d1 
	nullify(dd)
	if (associated(u)) dd => u
	do while (associated(d))
	  if (.not.singleton(d%s%name) .and. .not.indomain(d%s,u)) then
	    if (associated(u)) then
	      call newdomain(dd%next)
	      dd => dd%next
	    else
	      call newdomain(u)
	      nullify(dd)
	      dd => u
	    end if
	    dd%s => d%s
	  end if
	  d => d%next
	end do
	return
	end subroutine union

!!	Determine the subdomain of d1 in d2, i.e. those element of
!!	d1 which are not in d2:
!!>
!!                  subdomain = d1 \ (d1 ^ d2)
!!<
!!	The output is the set of elements in d1 which are not 
!!	singletons and not in d2.
!!
	subroutine findsubdomain(d1,d2,subdomain)
	type (domain), pointer :: d1, d2, subdomain, d, dd

	nullify(subdomain)

!	Look at each element in d1:

	d => d1 
	do while (associated(d))
	  if (.not.indomain(d%s,d2)) then
	    if (associated(subdomain)) then
	      call newdomain(dd%next)
	      dd => dd%next
	    else
	      call newdomain(subdomain)
	      dd => subdomain
	    end if
	    dd%s => d%s
	  end if
	  d => d%next
	end do
	return
	end subroutine findsubdomain

	logical function singleton(string)
	character (len=*) :: string
	singleton = ((string(1:1).eq.'"') .or. (string(1:1).eq."'"))
	return
	end function singleton

	subroutine writeparam(ident, index, d, string)
	character (len=*) :: ident, string
	type (domain), pointer :: d
	integer :: index
	character (len=255) :: domstr

	call writedomain(d, domstr)
	if (domstr.eq.' ') then
	  write(string,'(a,i3.3)') trim(ident), index
	else
	  write(string,'(a,i3.3,a)') trim(ident), index,'('//trim(domstr)//')'
	end if
	return
	end subroutine writeparam

	subroutine domainlabels(d,dl)
        type (domain), pointer :: d, dd	
	character (len=32) ::  dl(10)
	integer :: i
	dl = ' '
        dd => d
	i = 0
        do while (associated(dd))
	  i = i + 1
	  dl(i) = dd%s%name
          dd => dd%next
        end do
        return
        end subroutine domainlabels

	subroutine writedomain(d, string)
	character (len=*) :: string
	type (domain), pointer :: d, dd
	dd => d
	string = ' '
	do while (associated(dd))
	  if (string.eq.' ') then
	    string = dd%s%name
	  else
	    string = trim(string)//','//dd%s%name
	  end if
	  dd => dd%next
	end do
	return
	end subroutine writedomain

!!	Read through variable declarations

	subroutine readvar
	type (symbol), pointer :: s
	type (function), pointer :: fn
	type (field), pointer :: f
	type (record), pointer :: r
	character (len=9) :: vtype
	logical :: pfield
	integer :: fieldno

	pause 'Entering READVAR'
	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  write(*,'(a)') trim(f%label)
          if ( f%label.eq.'$model') modelname = f%data
	  if (	f%label.ne.'$sectors'   .and. f%label.ne.'$commodities' .and. &
		f%label.ne.'$consumers' .and. f%label.ne.'$auxiliary') then
	    fn => fn%next
	    cycle 
	  else
	    select case(f%label)
	      case ('$sectors') 
	        vtype = 'sector'
	      case ('$commodities')
	        vtype = 'commodity'
	      case ('$consumers')
	        vtype = 'consumer'
	      case ('$auxiliary')
	        vtype = 'auxiliary'
	    end select
	  end if

	  if (log) write(*,'(a)') ' ** '//trim(f%label)

	  if (r%nfield.ne.1)  stop 'Error: error in variable declaration definition record.'
	  f => r%firstfield
	  if (f%data.ne.' ') stop 'Error: trailing items not permitted with variable declaration title.'

	  do while (associated(r%next))
	    r => r%next
	    f => r%firstfield
	    if (r%nfield.ne.1)  stop 'Error: only one field per record in declarations.'
	    if (f%label.ne.' ') stop 'Error: labels are not expected in declaration recors.'
	    write(*,'(a)')  vtype//' '//trim(f%data)
	    call findid(f%identtext, s)
	    if (associated(s)) stop 'Error: variable identifier has appeared previously.'
	    call addid(f%identtext, vtype, s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end do
	  fn => fn%next
	end do
	return
	end subroutine readvar

!!	Read through records for each of the functions and add symbols
!!	to the master list as they are encountered.  Check that the 
!!	input data conforms to syntax rules.

	subroutine chkprod

	type (symbol), pointer :: s
	type (function), pointer :: fn
	type (field), pointer :: f
	type (record), pointer :: r
	logical :: pfield
	integer :: fieldno


	fn => functionlist

	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  if (f%label.ne.'$prod') then
	    fn => fn%next
	    cycle 
	  end if
	  if (log) write(*,'(a)') ' ** '//trim(f%data)

!	Skip the elasticity nests on the first pass.


	  do while (associated(r%next))

	    r => r%next
	    pfield = .false.	! Flag to assure only one P: field per record
	    f => r%firstfield
	    fieldno = 1

	    do while (associated(f))

	      if (log) write(*,'(a)') ' -- '//trim(f%label)//':'//trim(f%data)

	      call chkfield(f,fieldno,pfield)

	      fieldno = fieldno + 1
	      f => f%next
	    end do
	  end do
	  fn => fn%next
	end do
	return
	end subroutine chkprod

!!	Evaluate fields in a production block record:
!!	- identifier dimension and type must be consistent
!!	- i: and o: must only appear as leading fields and reference prices
!!	- q: field must be second and reference a parameter 
!!	- p: fields may appear only once and must reference parameter
!!	- a: field must reference a consumer
!!	- n: field must reference an auxiliary variable 
!!	- m: field must reference a parameter
!!	- t: field must reference a parameter
!!
	subroutine chkfield(f,fieldno,pfield)
	integer :: fieldno
	logical :: pfield
	type (field), pointer :: f
	type (symbol), pointer :: s

!	We are scanning the fields assoiciated with named labels, hence we can skip
!	the nest-assignment fields:

	if (f%data.eq.' ' .or. f%data(1:1).eq.'$') then
	  if (index(f%label,'.tl').gt.0) stop ' *** Error: .tl syntax is not yet implemented.'
	  if (index(f%label,'.te').gt.0) stop ' *** Error: .te syntax is not supported.'
	  return
	endif

	select case (f%label)	! Evaluate fields which are not nesting assignments

	case('i')
	  if (fieldno.ne.1) stop ' *** Error: I: must be the leading field.'
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'commodity') stop ' *** Error: I: field must reference a commodity price'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of commodity price is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'commodity', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('o')
	  if (fieldno.ne.1) stop ' *** Error: O: must be the leading field.'
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'commodity') stop ' *** Error: O: field must reference a commodity price'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of commodity price is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'commodity', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('q')
	  if (fieldno.ne.2) stop ' *** Error: Q: must be in the second field.'
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'parameter') stop ' *** Error: Q: field must reference a parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('p')
	  if (pfield) stop ' *** Error: P: field appears more than once.'
	  pfield = .true.
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext, s)
	  if (associated(s)) then 
	    if (s%type.ne.'parameter') stop ' *** Error: P: field must reference a parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('a')
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	    if (s%type.ne.'consumer') stop ' *** Error: A: field must reference a consumer'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'consumer', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if


	case('n')
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	    if (s%type.ne.'auxiliary') stop ' *** Error: N: field must reference an AUXILIARY'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of AUXILIARY variable changes.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'auxiliary', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('m')
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	  if (s%type.ne.'parameter') stop ' *** Error: M: multiplier must reference an parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter variable changes.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case('t')
	  if (f%identtext(1:1).eq.'(') return
	  if (numeric(f%data)) return
	  call findid(f%identtext,s)
	  if (associated(s)) then 
	    if (s%type.ne.'parameter') stop ' *** Error: T: multiplier must reference an parameter'
	    if (s%dim.eq.-1) then
	      s%dim = dimension(f%arg)
	    else
	      if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter variable changes.'
	    end if
	    f%ident => s
	  else
	    call addid(f%identtext, 'parameter', s, f%line)
	    f%ident => s
	    s%dim = dimension(f%arg)
	  end if

	case default
	  stop ' *** Error: unrecognized field label in $prod block'

	end select 

	return

	end subroutine chkfield

!!	Read nest assignments from the first record of each $prod block

	subroutine readnest

	type(function), pointer :: fn
	type(record), pointer :: r
	type(field), pointer :: f
	type(nest), pointer ::	esub, &	! Top level input nest
				etrn, &	! Top level output nest
				n, &	! Last nest allocated
				nn,np	! Nest pointer

	logical :: subnest	! Logical flag that a subnest has been specified.
	integer :: fieldno
	character (len=255) :: nestid

!	First pass to set up nesting pointers:

	fn => functionlist
	nullify(f)
	nullify(n)
	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield

!	Set up the top level input and output nests:

	  call newnest(esub)
	  fn%inputs  => esub
	  esub%value = '0'
	  esub%ident = 's'

	  call newnest(etrn)
	  fn%outputs => etrn
	  etrn%value = '0'
	  etrn%ident = 't'
	  etrn%output = .true.

	  esub%next => etrn
	  n => etrn
	  subnest = .false.
	  fieldno = 1

!	Look through all the fields on this record following the first:

	  do while (associated(f%next))
	    f => f%next
	    fieldno = fieldno + 1
	    if (f%spantext.ne.' ') stop ' *** Span notation "#" not permited in elasticity.'

!	If an elasticity is specified as a parameter, verify that the parameter domain
!	is a subset of the function domain:

	    if ((f%argtext.ne.' ') .and. (.not.subset(f%arg, r%firstfield%arg))) then
    	      write(*,'(a)') ' '//trim(r%firstfield%label)//':'//trim(r%firstfield%data)//' => '//trim(r%firstfield%argtext)
    	      write(*,'(a)') ' '//trim(f%label)//':'//trim(f%data)//' => '//trim(f%argtext)
	      stop ' ***  Invalid elasticity field domain.'
	    end if

!	Nest ID is specfied in the field label:

    	    nestid = f%label
	    if (index(nestid,'.tl').gt.0) stop ' *** Error: .tl syntax is not yet implemented.'
	    if (index(nestid,'.te').gt.0) stop ' *** Error: .te syntax is not supported.'

	    if (nestid.eq.'s') then
	      if (subnest) stop ' *** s: field must preceed subnest fields.'
	      if (associated(esub%f)) stop ' *** Multiple s: fields in $prod block'
	      if (fieldno.gt.3)       stop ' *** s: field must be first or second elasticity'
	      esub%f => f
	      esub%value = f%data
	      cycle
	    endif

	    if (nestid.eq.'t') then
	      if (subnest) stop ' *** t: field must preceed subnest fields.'
	      if (associated(etrn%f)) stop ' *** Multiple t: fields in $prod block'
	      if (fieldno.gt.3)       stop ' *** t: field must be first or second elasticity'
	      etrn%f => f
	      etrn%value = f%data
	      cycle
	    endif

	    subnest = .true.
	    call newnest(n%next)
	    n => n%next

	    call parsenestid(nestid,n%ident,n%parentid)
	    n%f => f
	    n%value = f%data
	  end do

!	Now install pointers to parents where an explicit parent has
!	been declared:

	  n => fn%inputs
	  do while (associated(n))
	    if (n%parentid.eq.' ') then
	      if (.not.(associated(n,fn%inputs) .or. associated(n,fn%outputs))) n%level = 1
	      n => n%next
	      cycle
	    end if

!	An explicit parent has been declared - find it

	    call findnest(n%parentid, fn, np)
	    n%parent => np
	    n%level = np%level + 1

!	Add this nest to the list of subnests in p:

	    call addsubnest(n, np)
	    n => n%next
	  end do 
	    
          if (log .and. associated(fn%inputs)) then
	    write(*,'(a)') ' Nests:'
	    n => fn%inputs
	    do while (associated(n))
	      if (associated(n%parent)) write(*,'(a)') ' '//trim(n%ident)//' -> '//trim(n%parentid)
	      n => n%next
	    end do
	  end if

	  fn => fn%next
	end do

	return
	end subroutine readnest

!!	Split a nest elasticity field into a nest identifier and a 
!!	parent identifier:

	subroutine parsenestid(nestid,ident,parentid)
	character (len=*) :: nestid	!! Identifier for nest
	character (len=*) :: ident	!! Identifier of the nest itself
	character (len=*) :: parentid	!! Identifier of the nest parent 

	integer :: ic, jc
	ic = index(nestid,'(')
	jc = index(nestid,')',back=.true.)
	if (ic.gt.0 .and. jc.eq.0) then
	  write(*,'(a)') trim(nestid)
	  stop ' Invalid parent node -- missing closing parenthesis.'
	end if
	if (ic.eq.0 .and. jc.gt.0) then
	  write(*,'(a)') trim(nestid)
	  stop ' Invalid parent node -- missing opening parenthesis.'
	end if
	if (ic.eq.0) then
	  ident = nestid
	  parentid = ' '
	else
	  if (jc.lt.ic+2) then
	    write(*,'(a)') trim(nestid)
	    stop ' Invalid parent node -- parenthesis mismatch.'
	  end if
	  ident = nestid(1:ic-1)
	  parentid = nestid(ic+1:jc-1)
	end if
	return
	end subroutine parsenestid

!!	Read through the production function coefficients:
!!	- Create netput descriptions of each production input and output
!!	- Create linked list data structure characterizing the nested CES function

	subroutine readcoef
	type (function), pointer :: fn
	type (symbol), pointer :: s
	type (record), pointer :: r
	type (netput), pointer :: np
	type (domain), pointer :: d
	type (field), pointer :: f, fnf, taxagent, lastfield
	type (nest), pointer :: n
	logical :: input, conditional
	integer :: nparent

	nullify(taxagent)	! No tax agent defined at this point
	  
	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  fnf => r%firstfield
	  if (fnf%label.ne.'$prod') then
	    fn => fn%next
	    cycle
	  end if

! Extract field identifier and add it to the field record - added by KP (?omitted by TR?)
	  call findid(fnf%identtext, s)

! 29/01/04 This is probably unnecessary becase of addition of f%ident in routine parse
! Add this identifier s to the field record - added by KP (?omitted by TR?)
	  if(.not.(associated(fnf%ident))) then
	    write(*,*)'KP add - was not already asociated'
	    fnf%ident => s
	  else
	    write(*,*)'KP add - was already associated'
	  end if

!	Read all the records in this function after the elasticity record:

	  do while (associated(r%next))
	    r => r%next

!	Look through the fields in this record:

	    f => r%firstfield
	    nparent = 0
	    call findid(f%identtext, s)

!	Create a netput for this record:

	    call newnetput(np)
	    np%f => f
	    np%output = (f%label.eq.'o')

	    np%nestcount = '0'
	    np%nestfields = 0
	    conditional = .true.

	    if (associated(fnf%arg)) call union(fnf%arg, np%d)
	    if (associated(f%arg))   call union(f%arg,  np%d)
	    if (associated(f%span))  call union(f%span, np%d)
	    if (associated(np%d))    call findsubdomain(np%d, fnf%arg, np%sd)
	    nullify(lastfield)	! Reading the first field -- no prior field at this point


!	Look for nest identifiers:

	    do while (associated(f%next))
	      f => f%next

!	A nest identifier is a field with no data component:

	      if (f%data.eq.' ') then

!	Locate this nest identifier in the master list:

	        call findnest(f%label, fn, n)

!	Record the nest type:

		call nesttype(fn, n, np%output)

!	Add this netput to this nest:

		call addnetput(np, n, f)
		nullify(lastfield)
		np%nestfields = np%nestfields + 1
		if (f%condtn.ne.' ') then
		  np%nestcount = trim(np%nestcount)//'+1$'//trim(f%condtn)
		else
		  np%nestcount = trim(np%nestcount)//'+1'
		  conditional = .false.
		end if
	      else
	        call readelem(np, f, taxagent, lastfield)
		lastfield => f
	      end if
	    end do

!	Assign this netput to a toplevel

	    if (     np%output) n => fn%outputs
	    if (.not.np%output) n => fn%inputs
	    if (np%nestfields.eq.0) then
	      nullify(f)
	      call addnetput(np, n, f)
	    else
	      if (conditional) then
	        call newfield(f)
		f%condtn = '('//trim(np%nestcount)//'=0)'
		call addnetput(np, n, f)
	      endif
	    end if
	  end do
	  fn => fn%next
	end do

	return
	end subroutine readcoef

!!	Introduce type definitions for nests beginning at n
!!	and proceeding to the top of the tree.

	subroutine nesttype(fn, n, output)
	type (function), pointer :: fn	!! Function we fare looking at
	type (nest), pointer :: n	!! Nest from which to trace
	logical :: output		!! Flag for whether this is an output nest
	type (nest), pointer :: np
	if (n%output .and. .not.output) stop 'Error: input enters an output nest'
	n%output = output
	if (.not. associated(n%parent)) then
	  if (output) then
	    n%parent => fn%outputs
	  else
	    n%parent => fn%inputs
	  end if
	  return
	else
	  np => n
	  do while (associated(np%parent))
	    np => np%parent
	    if (np%output .and. .not.output) stop 'Error: input enters an output nest'
	    np%output = output
	  end do
	  if (output      .and. .not.associated(np,fn%outputs)) np%parent => fn%outputs
	  if (.not.output .and. .not.associated(np,fn%inputs))  np%parent => fn%inputs
	end if
	return
	end subroutine nesttype

!!	Read a single field and introduce the associated logic
!!	in a netput descriptor:

	subroutine readelem(np, f, taxagent, lastfield)
	type (netput), pointer :: np	!! Netput to which this should be assigned
	type (field), pointer  :: f	!! Field for which logic is to be introduced
	type (field), pointer :: taxagent	!! Field which defines applicable tax agent field
	type (field), pointer :: lastfield	!! Pointer to the previous field in this record (if one)
	type (taxlist), pointer :: t

	select case (f%label)

	case ('p')
	  if (associated(np%pref)) stop 'P: field appears more than once.'
	  np%pref => f

	case ('q')
	  if (associated(np%qref)) stop 'Q: field appears more than once.'
	  np%qref => f

	case ('a')
	  taxagent => f

	case ('t')
	  if (.not.associated(taxagent)) stop 'Tax field appears prior to specification of a tax agent (A:).'
	  call addtax(np,t)
	  t%t => f
	  t%a => taxagent

	case ('n')
	  if (.not.associated(taxagent)) stop 'Tax field appears prior to specification of a tax agent (A:).'
	  call addtax(np,t)
	  t%n => f
	  t%a => taxagent
	  
	case ('m')
	  if (.not.associated(taxagent)) stop 'Tax field appears prior to specification of a tax agent (A:).'
	  call lasttax(np,t)
	  if (.not.associated(lastfield))     stop 'M: field must follow N: field'
	  if (.not.associated(t%n,lastfield)) stop 'M: field must follow immediately after N: field'
	  t%m => f

	end select

	lastfield => f
	return
	end subroutine readelem

!!	Create symbol table entries for the production sectors

	subroutine sectors

	character (len=31) :: label
	type(function), pointer :: fn
	type(record), pointer :: r
	type(field), pointer :: f
	type(symbol), pointer :: s
	type(domain), pointer :: d

        fn => functionlist
        nullify(f);
        do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  if (log) write(*,'(a)') ' '//trim(f%label)//':'//trim(f%data)
	  fn => fn%next
        end do

	nullify(f);
	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  f => r%firstfield
	  if (log) write(*,'(a)') ' label: '//trim(f%label)

	  if (f%label.eq.'$prod') then
	    if (log) write(*,'(a)') ' Checking $prod for: '//trim(f%identtext)
	    call findid(f%identtext, s)
	    if (.not.associated(s)) then
	      write(*,'(a)') f%identtext
	      stop ' *** $prod block found for undeclared $sector.'
	    end if
	    if (s%dim.ne.dimension(f%arg)) stop '$prod variable dimension is inconsistent'
	    if (f%spantext.ne.' ') stop ' *** Span notation "#" not permited in $prod block.'
	    d => f%arg
	    do while (associated(d))
	      if (d%s%name(1:1).eq.'"' .or. d%s%name(1:1).eq."'") &
		stop ' *** Singleton sets not permitted in $prod domain'
	      d => d%next
	    end do
	  end if

          if (f%label.eq.'$demand') then
            if (log) write(*,'(a)') ' Checking $demand for '//trim(f%identtext)
            call findid(f%identtext, s)
            if (.not.associated(s)) then
              write(*,'(a)') f%identtext
              stop ' *** $demand block found for undeclared $consumer.'
            end if
            if (s%dim.ne.dimension(f%arg)) stop '$demand variable dimension is inconsistent'
            if (f%spantext.ne.' ') stop ' *** Span notation "#" not permited in $demand block.'
            d => f%arg
            do while (associated(d))
              if (d%s%name(1:1).eq.'"' .or. d%s%name(1:1).eq."'") &
                stop ' *** Singleton sets not permitted in $demand domain'
              d => d%next
            end do
          end if

	  fn => fn%next
	end do
	return
	end subroutine sectors

!!	Determine which a string is numeric:

	logical function numeric(string)
	character (len=*) :: string	!! String to be evaluated (.true. if a Fortran-readable numeric value)
	real :: testval
	integer :: stat
	if (index('.+-0123456789', string(1:1)).gt.0) then
	  read(string,*,iostat=stat) testval
	  if (stat.ne.0) stop ' ** Error in numeric field.'
	  numeric = .true.
	else
	  numeric = .false.
	end if
	return
	end function numeric

!!	Add a netput to a nest.

	subroutine addnetput(np, n, f)
	type (netput), pointer :: np	!! Netput to be added
	type (nest), pointer :: n	!! Nest in which the netput is to be introduced
	type (field), pointer :: f	!! Field in which the nest assignment is made (for condition if any)
	type (netputlist), pointer :: nl
	if (.not.associated(n%elements)) then
	  nullify(nl)
	  allocate(nl)
	  n%elements => nl
	else
	  nl => n%elements
	  do while (associated(nl%next))
	    nl => nl%next
	  end do
	  allocate(nl%next)
	  nl => nl%next
	endif
	nullify(nl%np)
	nullify(nl%next)
	nl%np => np
	if (associated(f)) nl%f => f
	return
	end subroutine addnetput

!!	Add a tax coefficient to the associated linked list for a production netput:

	subroutine addtax(np,t)
	type (netput), pointer :: np	!! Netput to which tax is applied
	type (taxlist), pointer :: t	!! Pointer to the tax field

	if (associated(np%t)) then
	  call lasttax(np, t)
	  call newtax(t%next)
	  t => t%next
	else
	  call newtax(np%t)
	  t => np%t
	end if
	return
	end subroutine addtax

!!	Set pointer to the last tax applied to a given netput:

	subroutine lasttax(np, t)
	type (netput), pointer :: np	!! Netput to be queried
	type (taxlist), pointer :: t	!! On output points to the last tax coefficient

	t => np%t
	do while (associated(t%next))
	  t => t%next
	end do
	return
	end subroutine lasttax


!!	Generate a pointer to the nest with a given name in a particular function:

	subroutine findnest(nestid, fn, n)
	character (len=*) :: nestid	!! String descriptor of the nest to be located
	type (function), pointer :: fn	!! Function in which to locate the nest
	type (nest), pointer :: n	!! Pointer to the located nest

	n => fn%inputs
	if (log) write(*,'(a)') ' Locating nest: '//trim(nestid)
	do while (associated(n))
	  if (nestid.eq.n%ident) return
	  n => n%next
	end do
	stop 'nestid not found'
	end subroutine findnest
	    
!!	At a subnest to a nest.  This requires creation of a nestlist if
!!	one does not already exist

	subroutine addsubnest(n, np)
	type (nest), pointer :: n	!! Nest to be added as a subnest
	type (nest), pointer :: np	!! Nest to which the subnest is to be introduced
	type (nestlist), pointer :: nl

	if (.not.associated(np%subnests)) then
	  nullify(nl)
	  allocate(nl)
	  np%subnests => nl
	  nullify(nl%next)
	  nl%n => n
          n%nlist => nl
	else
	  nl => np%subnests
	  do while (associated(nl%next))
	    nl => nl%next
	  end do
	  allocate(nl%next)
	  nl => nl%next
	  nullify(nl%next)
	  nl%n => n
          n%nlist => nl
	end if
	return
	end subroutine addsubnest

!!	Returns .true. if the argument is quoted text, i.e. a character
!!	string beginning and ending with a pair of single or double quotes.

	logical function quoted(arg)
	character (len=*) :: arg
	integer :: lc
	lc = len_trim(arg)
	quoted = .true.
	if (arg(1:1).eq."'" .and. arg(lc:lc).eq."'") return
	if (arg(1:1).eq.'"' .and. arg(lc:lc).eq.'"') return
	quoted = .false.
	return
	end function quoted

!!	Public function determines whether two strings are
!!	identical apart from case.  

        logical function eqv(t1,t2)
	implicit none
        integer :: i
        character (len=*) :: t1,t2
	character (len=255) :: s1,s2
	character (len=26) :: upper = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', &
			      lower = 'abcdefghijklmnopqrstuvwxyz'

	s1 = t1
	s2 = t2
	if (quoted(t1)) s1 = t1(2:len_trim(t1)-1)
	if (quoted(t2)) s2 = t2(2:len_trim(t2)-1)
	  
	eqv = .false.
	if (len_trim(s1).ne.len_trim(s2)) return
        do i=1,len_trim(s1)
          if (.not. ((s1(i:i).eq.s2(i:i)) .or. &
	      (index(upper,s1(i:i))+index(upper,s2(i:i)).ne.0 .and. &
	       index(upper,s1(i:i)).eq.index(lower,s2(i:i)) .and. &
	       index(upper,s2(i:i)).eq.index(lower,s1(i:i))))) return
	end do
	eqv = .true.
	return
	end function eqv

        subroutine consumers

        character (len=31) :: label
        type(function), pointer :: fn
        type(record), pointer :: r
        type(field), pointer :: f
        type(symbol), pointer :: s
        type(domain), pointer :: d

        nullify(f);
        fn => functionlist
        do while (associated(fn))
          r => fn%firstrecord
          f => r%firstfield
          fn => fn%next
        end do
        return
        end subroutine consumers


!!      Read through records for each of the demand functions and add symbols
!!      to the master list as they are encountered.  Check that the 
!!      input data conforms to syntax rules.

        subroutine chkdemand

        type (symbol), pointer :: s
        type (function), pointer :: fn
        type (field), pointer :: f,fnf
        type (record), pointer :: r
	type (endowmentlist), pointer :: el
	type (endowment), pointer :: e
        integer :: fieldno
	type (domain), pointer   :: d           ! Pointer to the domain of a D: field


        fn => functionlist
        do while (associated(fn))
          r => fn%firstrecord
          fnf => r%firstfield
          if (fnf%label.ne.'$demand') then
            fn => fn%next
            cycle 
          end if
	  f => r%firstfield
          if (log) write(*,'(a)') ' ** '//trim(f%data)

          do while (associated(r%next))
            r => r%next
            f => r%firstfield
            fieldno = 1
	    if (f%label.eq.'d') then
	      if (associated(fn%demand)) stop '*** Error: multiple D: fields encountered.'
	      fn%demand => f
	      if (associated(f%arg)) then
	        if (associated(fnf%arg)) then
	          if (.not.subset(f%arg,fnf%arg)) stop '*** Error: D: fields must refer to single commodities.'
		else
		  nullify(d)
		  call union(f%arg,d)
		  if (d%dim.gt.0) stop '*** Error: D: field must refer to single commodities.'
		end if
	      end if
	      if (r%nfield.gt.2)  stop '*** Error: D: fields must be singleton fields, one per consumer.'
	    end if

	    if (f%label.eq.'e') then
	      if (.not.associated(fn%endowments)) then
	        allocate(el)	      
		fn%endowments => el
	      else
	        allocate(el%next)
		el => el%next
	      endif
	      nullify(el%next)
	      allocate(e)
	      nullify(e%q);nullify(e%r);nullify(e%d);nullify(e%sd)
	      el%e => e
	      e%f => f
	      if (associated(fnf%arg)) call union(fnf%arg, e%d)
	      if (associated(f%arg))   call union(f%arg,   e%d)
	      if (associated(f%span))  call union(f%span,  e%d)
	      if (associated(e%d))     call findsubdomain(e%d, fnf%arg, e%sd)
	    else
	      nullify(e)
	    end if
            do while (associated(f))
              if (log) write(*,'(a)') ' -- '//trim(f%label)//':'//trim(f%data)
              call chkdfield(f,fieldno,e)
              fieldno = fieldno + 1
              f => f%next
            end do
          end do
	  if (.not.associated(fn%demand)) stop '*** Error: require exactly one D: field per consumer.'
          fn => fn%next
        end do
        return
        end subroutine chkdemand

!!      Evaluate fields in a demand block record:
!!      - identifier dimension and type must be consistent
!!      - e: and d: must only appear as leading fields 
!!      - q: field must be second and reference a parameter 
!!      - r: field must reference an auxiliary and be the third
!!
        subroutine chkdfield(f,fieldno,e)
        integer :: fieldno
        type (field), pointer :: f
        type (symbol), pointer :: s
	type (endowment), pointer :: e

        select case (f%label)   ! Evaluate fields which are not nesting assignments

        case('d')
          if (fieldno.ne.1) stop ' *** Error: D: must be the leading field.'
          call findid(f%identtext, s)
          if (associated(s)) then 
            if (s%type.ne.'commodity') stop ' *** Error: D: field must reference a commodity price'
            if (s%dim.eq.-1) then
              s%dim = dimension(f%arg)
            else
              if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of commodity price is different.'
            end if
            f%ident => s
          else
            call addid(f%identtext, 'commodity', s, f%line)
            f%ident => s
            s%dim = dimension(f%arg)
          end if

        case('e')
          if (fieldno.ne.1) stop ' *** Error: E: must be the leading field.'
          call findid(f%identtext, s)
          if (associated(s)) then 
            if (s%type.ne.'commodity') stop ' *** Error: E: field must reference a commodity price'
            if (s%dim.eq.-1) then
              s%dim = dimension(f%arg)
            else
              if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of commodity price is different.'
            end if
            f%ident => s
          else
            call addid(f%identtext, 'commodity', s, f%line)
            f%ident => s
            s%dim = dimension(f%arg)
          end if

        case('q')
	  if (.not.associated(e)) stop '*** Error: Q: field only expected in an endowment record with a $demand block'
          if (fieldno.ne.2) stop ' *** Error: Q: must be in the second field.'
	  e%q => f
          if (f%identtext(1:1).eq.'(') return
          if (numeric(f%data)) return
          call findid(f%identtext, s)
          if (associated(s)) then 
            if (s%type.ne.'parameter') stop ' *** Error: Q: field must reference a parameter'
            if (s%dim.eq.-1) then
              s%dim = dimension(f%arg)
            else
              if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of parameter is different.'
            end if
            f%ident => s
          else
            call addid(f%identtext, 'parameter', s, f%line)
            f%ident => s
            s%dim = dimension(f%arg)
          end if

        case('r')
	  if (.not.associated(e)) stop '*** Error: R: field only expected in an endowment record with a $demand block'
	  if (associated(e%r)) stop '*** Error: multiple R: fields?'
	  e%r => f
          if (f%identtext(1:1).eq.'(')	stop ' *** Error: R: field must reference an auxiliary'
          if (numeric(f%data))		stop ' *** Error: R: field must reference an auxiliary'
          call findid(f%identtext, s)
          if (associated(s)) then 
            if (s%type.ne.'auxiliary') stop ' *** Error: R: field must reference an auxiliary'
            if (s%dim.eq.-1) then
              s%dim = dimension(f%arg)
            else
              if (s%dim.ne.dimension(f%arg))  stop ' *** Error: dimension of auxiliary is different.'
            end if
            f%ident => s
          else
            call addid(f%identtext, 'auxiliary', s, f%line)
            f%ident => s
            s%dim = dimension(f%arg)
          end if

        case default
          stop ' *** Error: unrecognized field label in $prod block'

        end select 

        return

        end subroutine chkdfield

! Subroutines added by KP begin here

	subroutine ConvertToGP
!  Does declarations for $prod
!  Based on Tom Rutherford's routine "declarations"
	type (field), pointer :: fnf
	type (record), pointer :: r
	type (function), pointer :: fn
	type (nest), pointer :: n
	type (netputlist), pointer :: e
	type (nest), pointer :: inputs, outputs
	type (domain), pointer :: InputsSubDomain
! for $prod
	type (GPCoeffVarReal) :: InputPrice, InputQuantity 
	type (GPCoeffVarReal) :: OutputPrice, OutputQuantity 
	type (GPCoeffVarReal) :: ThisProdVar, ThisTopVar
	type (GPCoeffVarReal) :: ThisElasticity
	type (symbol), pointer :: ElasticitySymbol
! for $demand
	type (GPCoeffVarReal) :: DemandPrice, DemandQuantity 
	type (GPCoeffVarReal) :: EndowmentPrice, EndowmentQuantity
	type (GPCoeffVarReal) :: ThisDemandVar
	logical :: DFieldExists, EFieldExists 

	integer :: fnindex = 0
	character (len=1024) :: rec
	logical DoThis

10015   format(1x,a)     ! KP
10017   format(2x,3a)

! Common declarations
	call StartTABGP

	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  fnf => r%firstfield
	  if (fnf%label.eq.'$prod') then

! Example
! $prod:y(s)          s:sigma(s)         
!  o:p(s)            q:supply(s)
!  i:pf(f)           q:factor(f,s)

	    write(*,10015) 'Next $prod'
	    write(*,10017) '%data: ', trim(fnf%data)
	    write(*,10017) '%identtext: ', trim(fnf%identtext)
	    write(*,10017) '%argtext: ', trim(fnf%argtext)
! ??
	    call WriteRecord('$prod record:', r)
	    if(associated(fnf)) then
!	      call DeclareFieldGP('Variable', fnf)
	    end if
	    inputs => fn%inputs
	    outputs => fn%outputs
	    call WriteNest('INPUTS',inputs)
	    call WriteNest('OUTPUTS',outputs)
! Get input and output price and quantity
! In the example, InputPrice is pf(f), InputQuantity is factor(f)
	    call GetProdPriceQuantity(fn, 'i', InputPrice, InputQuantity)
! In the example, OutputPrice is p(s), OutputQuantity is supply(s)
	    call GetProdPriceQuantity(fn, 'o', OutputPrice, OutputQuantity)
! Put out comment above the declarations for this $prod
	    call GPCommentFunction(fn)
! Declare input and output quantities - these are coefficients
	    call DeclareGPCoeffVar(OutputQuantity) 
	    call DeclareGPCoeffVar(InputQuantity)
	    call ReadCoeffGP(InputQuantity)
! Get Variable ThisProdVar associated with the top level of this $prod function
! In the example, ThisProdVar is the variable on LHS of the CES equation qd_y_pf
	    call GetProdVariableGP(fn, ThisProdVar)
! Get input subdomain
! In the example, InputSubDomain consists of set f
	    call GetInputsSubDomain(fn, InputsSubDomain)
! Write the formula connecting these coefficients
	    call WriteProdQuantityFormula(OutputQuantity, InputQuantity, InputsSubDomain) 
! Get $prod variable and declare it
! In the example, ThisTopVar is y(s) and ThisElasticity is sigma(s)
	    call GetProdTopVariableElast(fn, ThisTopVar, ThisElasticity)
	    call DeclareGPCoeffVar(ThisTopVar)
! Declare elasticity if is a Coeff
	    if(ThisElasticity%type.eq.'c') then
!  First find out if it is already declared - if so do not declare or read it
	      DoThis = .true.
	      call findid(ThisElasticity%name, ElasticitySymbol)
	      if(associated(ElasticitySymbol)) then
		write(*,*)'Elasticity ', trim(ThisElasticity%name), ' has a symbol'
		if(ElasticitySymbol%DeclaredGP) then
		  DoThis = .false.
		end if
	      else
		write(*,*)'Elasticity ', trim(ThisElasticity%name), ' has no symbol'
	      end if
	      if (DoThis) then
! Ideally should declare this as a Coefficient(parameter)
	        call DeclareGPCoeffVar(ThisElasticity)
	        call ReadCoeffGP(ThisElasticity)
	      end if
	    end if 
! Declare input and output price and quantity
	    call DeclareGPCoeffVar(OutputPrice) 
	    call DeclareGPCoeffVar(InputPrice) 
! Declare Variable associated with the top level of this $prod function
	    call DeclareGPCoeffVar(ThisProdVar)
! Write update for InputQuantity - delay until now since need ThisProdVar to be declared first
	    call WriteProdUpdate(InputQuantity, InputPrice, ThisProdVar)
! Write demand equation associated with this $prod function
	    call WriteProdDemandEquation(fn) 
! Write zero profit equation for this $prod function
	    call WriteProdZeroProfitEquation( &
		ThisTopVar, &
		OutputQuantity, OutputPrice, &
		InputQuantity, InputPrice, InputsSubDomain)

!	   End "if (fnf%label.eq.'$prod') then"
	  else if (fnf%label.eq.'$demand') then

! Example
! $demand:ra
!  d:pc              q:cons
!  e:pf(f)           q:endow(f)

! inputs and outputs are irrelevant for $demand
!	    inputs => fn%inputs
!	    outputs => fn%outputs
!	    write(*,*)'Calling WriteNest(inputs) for this $demand'
!	    call WriteNest('INPUTS',inputs)
!	    write(*,*)'Calling WriteNest(outputs) for this $demand'
!	    call WriteNest('OUTPUTS',outputs)

	    write(*,*)'Calling WriteRecord for this $demand'
	    call WriteRecord('$demand record:', r)

! Put out comment above the declarations for this $demand
	    call GPCommentFunction(fn)

! Get demand and endowment price and quantity
! In the example, DemandPrice is pc, DemandQuantity is cons
! In the example, EndowmentPrice is pf(f), EndowmentQuantity is endow(f)
! Also get $demand variable ThisTopVar and declare it
! In the example, ThisTopVar is ra
	    call GetDemandPriceQuantities(fn, &
	      DemandPrice, DemandQuantity, DFieldExists, &
	      EndowmentPrice, EndowmentQuantity, EFieldExists, &
	      ThisTopVar)
! Declare input and output quantities - these are coefficients
	    if(DFieldExists) then
	      call DeclareGPCoeffVar(DemandQuantity)
	    end if 
	    if(EFieldExists) then
	      call DeclareGPCoeffVar(EndowmentQuantity)
! Read it
	      call ReadCoeffGP(EndowmentQuantity)
	    end if 
	    call DeclareGPCoeffVar(ThisTopVar)
! Declare the Demand and Endowment price variables
	    if(DFieldExists) then
	      call DeclareGPCoeffVar(DemandPrice)
	    end if
	    if(EFieldExists) then
	      call DeclareGPCoeffVar(EndowmentPrice)
	    end if
! Get Variable ThisDemandVar associated with the top level of this $prod function
! In the example, ThisDemandVar is the variable qs_ra_pf 
	    if(EFieldExists) then
	      call GetDemandVariableGP(ThisTopVar, EndowmentPrice, ThisDemandVar)
! Declare this variable
	      call DeclareGPCoeffVar(ThisDemandVar)
	    end if
! Write update for Endowment quantity
	    if(EFieldExists) then
	      call WriteDemandUpdate(ThisDemandVar, &
		EndowmentPrice, EndowmentQuantity)
	    end if
! Write the market clearing for this $demand block
	    call WriteDemandMarketClearingEquation(ThisTopVar, DemandPrice)
! Write the income balance equation for this $demand block
	    if(EFieldExists) then
	      call WriteDemandIncomeBalanceEquation(ThisTopVar, &
		EndowmentPrice, EndowmentQuantity, ThisDemandVar)
	    end if


!	   End "if (fnf%label.eq.'$prod') then..else if(fnf%label.eq.'$demand') then.."
	  end if
	  fn => fn%next
	end do

! Write the market clearing equations

	return
	end subroutine ConvertToGP

	subroutine StartTABGP

! Local
	type (symbol), pointer :: ThisSymbol
	character (len=4) HeadString
	character (len=80) TextLine, SetName

10005	format(a)

	write(iotabgp,10005)'File Input # contains data and set elements # ;'

! Read all sets

	ThisSymbol => idlist
	do while(associated(ThisSymbol))
	  if (ThisSymbol%type.eq.'set') then
! Declare this set
	    NHeaderGp = NHeaderGP + 1
	    write(HeadString, '(I4.4)') NHeaderGp
	    SetName = trim(ThisSymbol%name)
	    call upper(SetName)
	    TextLine='Set '//trim(SetName)//' Read Elements from File Input Header "'//HeadString//'" ;'
	    write(iotabgp,10005) trim(TextLine)
	  end if
	  ThisSymbol => ThisSymbol%next
	end do

	return
	end subroutine StartTABGP

	subroutine WriteProdQuantityFormula(OutputQuantity, InputQuantity, InputsSubDomain)
! Input
	type (GPCoeffVarReal) :: OutputQuantity, InputQuantity
	type (domain), pointer :: InputsSubDomain 
! Local
	character (len=512) OutputString, TempString
	type (domain), pointer :: ThisDomain
	integer :: NSum, i
	character (len=128) :: ThisIndex 

10005	format(a)

! Formula keyword and quantifiers
	call ArgToAllsGP(OutputQuantity%arg, TempString)
	OutputString = 'Formula '//trim(TempString)
	write(iotabgp, 10005) trim(OutputString)
! The LHS
	call CoeffVarToGP(OutputQuantity, TempString)
	OutputString = trim(TempString)//' = '
	write(iotabgp, 10005) '  '//trim(OutputString)
! The RHS  "SUM{<index>,<set>,<InputQuantity>} ;"
! 29/01/04 Use routine WriteSumsGP for the sums
	call WriteSumsGP(InputsSubDomain, OutputString, NSum)
! WAS
! WAS	NSum = 0
! WAS	ThisDomain => InputsSubDomain
! WAS	OutputString = ''
! WAS	do while(associated(ThisDomain))
! Sum over this set
! WAS	  call GPSet2Index(ThisDomain%s, ThisIndex)
! WAS	  TempString = 'SUM{'//trim(ThisIndex)//','//trim(ThisDomain%s%name)//','
! WAS	  OutputString = trim(OutputString)//trim(TempString)
! WAS	  NSum = NSum + 1
! WAS	  ThisDomain => ThisDomain%next
! WAS	end do

!   The Top variable
	call CoeffVarToGP(InputQuantity, TempString)
	OutputString = trim(OutputString)//trim(TempString)
! End the sums
	call EndSumsGP(NSum, TempString)
	OutputString = trim(OutputString)//trim(TempString)//' ;'
	write(iotabgp,10005)'   '//trim(OutputString)

	return
	end subroutine WriteProdQuantityFormula

	subroutine WriteSumsGP(ThisDomain, OutputString, NSum)
! Input
	type (domain), pointer :: ThisDomain
! Output
	character (len=*), intent(out) :: OutputString
	integer, intent(out) :: NSum
! Local
	type (domain), pointer :: CurrentDomain
	character (len=128) :: CurrentIndex, TempString 

! The RHS  "SUM{<index>,<set>,<InputQuantity>} ;"
	NSum = 0
	CurrentDomain => ThisDomain
	OutputString = ''
	do while(associated(CurrentDomain))
! Sum over this set
	  call GPSet2Index(CurrentDomain%s, CurrentIndex)
	  TempString = 'SUM{'//trim(CurrentIndex)//','//trim(CurrentDomain%s%name)//','
	  OutputString = trim(OutputString)//trim(TempString)
	  NSum = NSum + 1
	  CurrentDomain => CurrentDomain%next
	end do

	return
	end subroutine WriteSumsGP

	subroutine EndSumsGP(NSum, OutputString)
! Input
	integer, intent(in) :: NSum
! Output
	character (len=*), intent(out) :: OutputString
! Local
	integer :: i
 
	OutputString=''
	do i = 1, NSum
	  OutputString = trim(OutputString)//'}'
	end do

	return
	end subroutine EndSumsGP


	subroutine WriteProdDemandEquation(ThisProdFunction)
! Input
	type (function), pointer :: ThisProdFunction
! Local
	character (len=512) OutputString, TempString
	type (GPCoeffVarReal) :: InputPrice, InputQuantity 
	type (GPCoeffVarReal) :: OutputPrice, OutputQuantity 
	type (GPCoeffVarReal) :: ThisProdVar, ThisTopVar
	type (GPCoeffVarReal) :: ThisElasticity
	character (len=LengthGPDeclarationName) EquationName


	call GetProdTopVariableElast(ThisProdFunction, ThisTopVar, ThisElasticity)
! Get input and output price and quantity
	call GetProdPriceQuantity(ThisProdFunction, 'i', InputPrice, InputQuantity)
	call GetProdPriceQuantity(ThisProdFunction, 'o', OutputPrice, OutputQuantity)
! Get Variable associated with the top level of this $prod function
	call GetProdVariableGP(ThisProdFunction, ThisProdVar)

10015	format(a)

! Equation keyword and equation name
	EquationName = 'E_'//trim(ThisProdVar%name)
	OutputString = 'Equation '//trim(EquationName)
	write(iotabgp, 10015) trim(OutputString)
! Alls
	call ArgToAllsGP(ThisProdVar%arg, OutputString)
	write(iotabgp, 10015) ' '//trim(OutputString)
! The equation
! The LHS
	call CoeffVarToGP(ThisProdVar, TempString)
	OutputString = trim(TempString)//' = '
	write(iotabgp, 10015) '  '//trim(OutputString)
! The RHS  "<top-variable> - "
!   The Top variable
	call CoeffVarToGP(ThisTopVar, TempString)
	OutputString = trim(TempString)//' - '
! Need elasticity 
! "* [ <input-price> - <output-price>]"
	if(ThisElasticity%type.eq.'r') then
	  OutputString = trim(OutputString)//' '//trim(ThisElasticity%name)//' * ['
	else
	  call CoeffVarToGP(ThisElasticity, TempString)
	  OutputString = trim(OutputString)//' '//trim(TempString)//' * ['
	end if
	call CoeffVarToGP(InputPrice, TempString)
	OutputString = trim(OutputString)//trim(TempString)//' - '
	call CoeffVarToGP(OutputPrice, TempString)
	OutputString = trim(OutputString)//trim(TempString)//'] ;'
	write(iotabgp, 10015) '   '//trim(OutputString)

! Add this declaration to list being built up
	call AddGPDeclaration('eq', trim(EquationName))

	return
	end subroutine WriteProdDemandEquation

	subroutine WriteProdZeroProfitEquation( &
	   ProdDemandVariable, &
           OutputQuantity, OutputPrice, &
           InputQuantity, InputPrice, InputsSubDomain)
! Input
	type (GPCoeffVarReal) :: ProdDemandVariable
	type (GPCoeffVarReal) :: OutputQuantity, InputQuantity
	type (GPCoeffVarReal) :: OutputPrice, InputPrice
	type (domain), pointer :: InputsSubDomain 
! Local
	character (len=512) OutputString, TempString, TempString2
	type (domain), pointer :: ThisDomain
	integer :: NSum, i
	character (len=128) :: ThisIndex 
	character (len=LengthGPDeclarationName) :: EquationName 

10005	format(a)

! Equation name
	EquationName='E_'//trim(ProdDemandVariable%name)
	OutputString = 'Equation '//trim(EquationName)
	OutputString = trim(OutputString)// &
	  '  # Zero profit for $prod:'// &
           trim(ProdDemandVariable%name)//' #'
	write(iotabgp, 10005) trim(OutputString)

! quantifiers
	call ArgToAllsGP(ProdDemandVariable%arg, TempString)
	if(TempString.ne.'') then
	  OutputString = ' '//trim(TempString)
	  write(iotabgp, 10005) trim(OutputString)
	end if
! The LHS
!  <OutputQuantity>*<OutputPrice> =
	call CoeffVarToGP(OutputQuantity, TempString)
	call CoeffVarToGP(OutputPrice, TempString2)
	OutputString = trim(TempString)//'*'//trim(TempString2)//' ='
	write(iotabgp, 10005) '  '//trim(OutputString)
! The RHS  "SUM{<index>,<set>,<InputQuantity>*<InputPrice>} ;"
! 29/01/04 Use routine WriteSumsGP for the sums
	call WriteSumsGP(InputsSubDomain, OutputString, NSum)
! WAS
! WAS	NSum = 0
! WAS	ThisDomain => InputsSubDomain
! WAS	OutputString = ''
! WAS	do while(associated(ThisDomain))
! Sum over this set
! WAS	  call GPSet2Index(ThisDomain%s, ThisIndex)
! WAS	  TempString = 'SUM{'//trim(ThisIndex)//','//trim(ThisDomain%s%name)//','
! WAS	  OutputString = trim(OutputString)//trim(TempString)
! WAS	  NSum = NSum + 1
! WAS	  ThisDomain => ThisDomain%next
! WAS	end do

	call CoeffVarToGP(InputQuantity, TempString)
	call CoeffVarToGP(InputPrice, TempString2)
	OutputString = trim(OutputString)//trim(TempString)//'*'// &
	  trim(TempString2)
! End the sums
	call EndSumsGP(NSum, TempString)
	OutputString = trim(OutputString)//trim(TempString)//' ;'
	write(iotabgp,10005)'   '//trim(OutputString)

	return
	end subroutine WriteProdZeroProfitEquation

	subroutine WriteProdUpdate(InputQuantity, InputPrice, ThisProdVar)
! Input
	type (GPCoeffVarReal), intent(in) :: ThisProdVar
	type (GPCoeffVarReal), intent(in) :: InputPrice, InputQuantity
! Local
	character (len=512) OutputString, TempString, TempString2

10005	format(a)

! quantifiers and LHS =
	call ArgToAllsGP(InputQuantity%arg, TempString)
	call CoeffVarToGP(InputQuantity, TempString2)
	OutputString='Update '//trim(TempString)//' '// &
	  trim(TempString2)//' ='
	write(iotabgp, 10005) trim(OutputString)

! RHS <InputPrice>*<ThisProdVar>
	call CoeffVarToGP(InputPrice, TempString)
	call CoeffVarToGP(ThisProdVar, TempString2)
	OutputString = '  '//trim(TempString)//' * '//trim(TempString2)//' ;'
	write(iotabgp, 10005) trim(OutputString)

	return
	end subroutine WriteProdUpdate

	subroutine GPCommentFunction(ThisFunction)
! Input
	type (function), pointer :: ThisFunction
! Local
	type (GPCoeffVarReal) :: ThisTopVar
	type (GPCoeffVarReal) :: ThisElasticity
	character (len=128) :: TempString
	type (record), pointer :: CurrentRecord
	type (line), pointer :: CurrentLine
	integer NRecords, KRecord

10015	format(a)

!	call GetProdTopVariableElast(ThisFunction, ThisTopVar, ThisElasticity)
! 	write(iotabgp,10015) '! $prod:'//trim(ThisTopVar%name)//' !'
	TempString = trim(ThisFunction%firstrecord%firstfield%data)
! Blank line
 	write(iotabgp,10015) ' '
 	write(iotabgp,10015) '! $prod:'//trim(TempString)//' !'
	NRecords = ThisFunction%nrecord
	KRecord = 0
! Write out the records as a strong comment
 	write(iotabgp,10015) '![[!'
	CurrentRecord => ThisFunction%firstrecord
	do while(associated(CurrentRecord))
	  CurrentLine => CurrentRecord%line
	  do while(associated(CurrentLine).and.KRecord.lt.NRecords)
	    write(iotabgp,10015) trim(CurrentLine%text)
	    KRecord=KRecord + 1
	    CurrentLine => CurrentLine%next
	  end do	  
	  CurrentRecord => CurrentRecord%next
	end do
 	write(iotabgp,10015) '!]]!'
	

	return
	end subroutine GPCommentFunction

	subroutine NewGPCoeffVar(ThisCoeffVar)
! Output
	type (GPCoeffVarReal) :: ThisCoeffVar

	ThisCoeffVar%type = ''    ! impossible type
	ThisCoeffVar%name = ''    ! impossible name
	nullify(ThisCoeffVar%arg)

	return
	end subroutine NewGPCoeffVar

	logical function IsReal(ThisString)
! Input
	character (len=*) ThisString
! Local
	integer LengthString, i
	character (len=12) RealChar
	character ThisChar
	intrinsic Index

	RealChar='0123456789.-'

	LengthString = LEN(trim(ThisString))
	IsReal = .true .
	do i = 1, LengthString
	  ThisChar = ThisString(i:i)
	  if(Index(RealChar,ThisChar).eq.0) then
	    IsReal = .false.
	  end if
	end do

	return
	end function IsReal

! Laurent calls this variable something like "overall production level variable"
	subroutine GetProdTopVariableElast(ThisFunction, ThisVariable, ThisElasticity)
! Gets variable after $prod: or after $demand:
! Input
	type (function), pointer :: ThisFunction
! Output
	type (GPCoeffVarReal) :: ThisVariable
	type (GPCoeffVarReal) :: ThisElasticity
! Local
	type (record), pointer :: r
	type (field), pointer :: fnf
	type (symbol), pointer :: ElastSymbol
	character (len=512) ElastText

	r => ThisFunction%firstrecord
	fnf => r%firstfield
	ThisVariable%type = 'v'
	ThisVariable%name = fnf%ident%name
	ThisVariable%arg => fnf%arg

! The elasticity

! Default (if not present) is zero, I presume
	ThisElasticity%type = 'r'
	ThisElasticity%name = '0'
	nullify(ThisElasticity%arg)
	
	fnf => fnf%next
	do while(associated(fnf))
	  if (fnf%label.eq.'s') then
! Is this a real or a Coefficient ?
	    ElastText = trim(fnf%data)
	    write(*,*)'ElastText=', ElastText
	    if(IsReal(ElastText)) then
	      ThisElasticity%name = trim(fnf%data)
	    else
	      ThisElasticity%type = 'c'
	      write(*,*)'ThisElasticity%data=', fnf%data
	      if(associated(fnf%ident)) then
	        ThisElasticity%name = fnf%ident%name
	      else
		write(*,*)'ThisElasticity: fnf%ident not associated'
		ThisElasticity%name = fnf%identtext
! Try to add %ident for this field fnf
		call findid(trim(fnf%identtext), ElastSymbol)
		if(associated(ElastSymbol)) then
		  write(*,*)' Elasticity ident added for ', trim(fnf%identtext),' - KP'
		  fnf%ident => ElastSymbol
		end if 
	      end if
	      ThisElasticity%arg => fnf%arg
	    end if
	  end if
	  fnf => fnf%next
	end do

	return
	end subroutine GetProdTopVariableElast

	subroutine GetProdPriceQuantity(ThisFunction, IOType, ThisPrice, ThisQuantity)
! Input
	type (function), pointer :: ThisFunction
	character, intent(in) :: IOType   ! 'i' or 'o'
! Output
	type (GPCoeffVarReal) :: ThisPrice, ThisQuantity
! Local
	type (record), pointer :: r
	type (field), pointer :: fnf
	type (nest), pointer :: outputs, inputs
	type (netputlist), pointer :: elements
	type (netput), pointer :: ThisNetPut
	type (field), pointer :: IOField, qref, pref
	logical :: UseThis, UseThis1, UseThis2

!	write(*,*)' GetProdPriceQuantity: IOType=', IOType
	r => ThisFunction%firstrecord
	fnf => r%firstfield
	if(IOType.eq.'o') then
	  outputs => ThisFunction%outputs
	  elements => outputs%elements
	else if(IOType.eq.'i') then
	  inputs => ThisFunction%inputs
	  elements => inputs%elements
!	else
!	  error
	end if
	do while (associated(elements)) 
	  ThisNetPut => elements%np
	  IOField => ThisNetPut%f   ! I: or O: field
	  qref => ThisNetPut%qref
	  pref => ThisNetPut%pref
	  UseThis1 = (IOType.eq.'o' .and. ThisNetPut%output)  
	  UseThis2 = (IOType.eq.'i' .and. .not.ThisNetPut%output)
	  UseThis = UseThis1.or.UseThis2  
	  if (UseThis) then
	    ThisPrice%type = 'v' ! Variable
	    ThisPrice%name = IOField%ident%name
	    ThisPrice%arg => IOField%arg	    
	    ThisQuantity%type = 'c' ! Coefficient
	    ThisQuantity%name = qref%ident%name
	    ThisQuantity%arg => qref%arg	    
!  Are done - return
	    return
	  end if
	  elements => elements%next
	end do

	return
	end subroutine GetProdPriceQuantity

	subroutine GetInputsSubDomain(ThisFunction, ThisSubdomain)
! Input
	type (function), pointer :: ThisFunction
! Output
	type (domain), pointer :: ThisSubdomain
! Local
	type (record), pointer :: r
	type (field), pointer :: fnf
	type (nest), pointer :: outputs, inputs
	type (netputlist), pointer :: elements
	type (netput), pointer :: ThisNetput
	logical :: DoneOk

	DoneOk = .false .
	r => ThisFunction%firstrecord
	fnf => r%firstfield
	inputs => ThisFunction%inputs
	elements => inputs%elements
	if (associated(elements)) then 
	  ThisNetput => elements%np
	  if(associated(ThisNetput)) then
	    ThisSubdomain => ThisNetput%sd
	    DoneOk = .true .
	  end if
	end if

	if(.not.DoneOk) then
	  write(*,*)'GetInputsSubDomain - not done ok'
	end if

	return
	end subroutine GetInputsSubDomain

	subroutine ReadCoeffGP(ThisCoefficient)
! Input
	type (GPCoeffVarReal) :: ThisCoefficient
! Local
	character (len=4) :: HeadString
	character (len=80) :: TextLine, CoeffName 	

10005	format(a)

	NHeaderGp = NHeaderGP+1
	write(HeadString, '(I4.4)') NHeaderGp
	CoeffName=trim(ThisCoefficient%name)
	call upper(CoeffName)
	TextLine = 'Read '//trim(CoeffName)//' from File Input Header "'//HeadString//'" ;'
	write(iotabgp, 10005) trim(TextLine)

	return
	end subroutine ReadCoeffGP	


	subroutine DeclareGPCoeffVar(ThisGPCoeffVar)
! Input
	type (GPCoeffVarReal), intent(in) :: ThisGPCoeffVar
! Local
	type (field), pointer :: ThisField
	type (symbol), pointer :: ThisSymbol
	character (len=16) CoeffVarString

	call newfield(ThisField)
	call newsymbol(ThisSymbol)
	ThisSymbol%name = ThisGPCoeffVar%name
	ThisField%ident => ThisSymbol
	ThisField%arg => ThisGPCoeffVar%arg
	if(ThisGPCoeffVar%type.eq.'c') then
	  CoeffVarString = 'Coefficient'
	else
	  CoeffVarString = 'Variable'
	end if

	call DeclareFieldGP(CoeffVarString, ThisField)
 
	return
	end subroutine DeclareGPCoeffVar	

	subroutine GetProdVariableGP(ThisProdFunction, ThisVar)
! Input
	type (function), pointer :: ThisProdFunction
! Output
	type (GPCoeffVarReal) :: ThisVar
! Local
	type (GPCoeffVarReal) :: InputPrice, InputQuantity 
	type (GPCoeffVarReal) :: OutputPrice, OutputQuantity 
	type (GPCoeffVarReal) :: ThisProdVar, ThisTopVar
	type (GPCoeffVarReal) :: ThisElasticity

	type (nest), pointer :: ProdInputs
	type (record), pointer :: ProdFirstRecord
	type (field), pointer :: ProdInputsField
	type (field), pointer :: ProdFirstField, CurrentNetputField
	type (symbol), pointer :: ProdIdent, CurrentNetputSymbol
	type (netputlist), pointer :: CurrentNetputList
	type (netput), pointer :: CurrentNetput
	type (domain), pointer :: CurrentNetputDomain

! Get $prod variable
	call GetProdTopVariableElast(ThisProdFunction, ThisTopVar, ThisElasticity)
! Get input and output price and quantity
	call GetProdPriceQuantity(ThisProdFunction, 'i', InputPrice, InputQuantity)
	ThisVar%name = 'qd_'//trim(ThisTopVar%name)//'_'//trim(InputPrice%name) 

	ProdFirstRecord => ThisProdFunction%firstrecord
	ProdFirstField => ProdFirstRecord%firstfield
	ProdIdent => ProdFirstField%ident

	ProdInputs => ThisProdFunction%inputs
	ProdInputsField => ProdInputs%f

	CurrentNetputList => ProdInputs%elements
	do while (associated(CurrentNetputList))
	  CurrentNetput => CurrentNetputList%np
	  CurrentNetputField => CurrentNetput%f
	  CurrentNetputSymbol => CurrentNetputField%ident
	  CurrentNetputDomain => CurrentNetput%d

	  ThisVar%type = 'v'

! The demand variable associated with this nest ranges over CurrentNetputDomain
!  as does the assoicated equation
! Following Laurent's convention, variable name is "qd_yy_pp" where the 
!  $prod field is followed by "yy" and the i field is followed by "pp"
! "yy" is ProdIdent%name - also ThisTopVar%name
! "pp" is CurrentNetputSymbol%name - also InputPrice%name

	  ThisVar%name = 'qd_'//trim(ThisTopVar%name)//'_'//trim(InputPrice%name) 
!	  ThisVar%name = 'qd_'//trim(ProdIdent%name)//'_'//trim(CurrentNetputSymbol%name)
 
	  ThisVar%arg => CurrentNetPutDomain
! Only doing this for single level function so Return here
	  Return
! Next would happen if processing several level $prod function
	  CurrentNetputList => CurrentNetputList%next 
	end do

	return
	end subroutine GetProdVariableGP

	subroutine GetDemandPriceQuantities(ThisDemandFunction, &
	      DemandPrice, DemandQuantity, DFieldExists, &
	      EndowmentPrice, EndowmentQuantity, EFieldExists, &
	      DemandTopVariable)
! Input
	type (function), pointer :: ThisDemandFunction
! Output
	type (GPCoeffVarReal), intent(out) :: DemandPrice, DemandQuantity
	type (GPCoeffVarReal), intent(out) :: EndowmentPrice, EndowmentQuantity
	type (GPCoeffVarReal), intent(out) :: DemandTopVariable
	logical :: DFieldExists, EFieldExists
! Local
	type (record), pointer :: ThisRecord
	type (record), pointer :: CurrentRecord
	type (field), pointer :: CurrentField
	character(len=255) CurrentLabel, LastLabel

	DFieldExists = .false .
	EFieldExists = .false .

	CurrentRecord => ThisDemandFunction%firstrecord
	do while(associated(CurrentRecord))
	  write(*,*)'GetDemandPriceQuantities - a new record'
	  CurrentField => CurrentRecord%firstfield
	  do while(associated(CurrentField))
	    CurrentLabel = CurrentField%label
	    call lower(CurrentLabel)

	    if(CurrentLabel.eq.'$demand') then
!  DemandTopVariable
	      DemandTopVariable%type = 'v'
	      if(associated(CurrentField%ident)) then
		write(*,*)'GetDemandPriceQuantities: DemandTopVariable%ident is associated'
	        DemandTopVariable%name = CurrentField%ident%name
	      else
		write(*,*)'GetDemandPriceQuantities: DemandTopVariable%ident is not associated'
	        DemandTopVariable%name = CurrentField%identtext
	      end if
	      DemandTopVariable%arg => CurrentField%arg
	    else if(CurrentLabel.eq.'d') then
! DemandPrice
	      DemandPrice%type = 'v'
	      if(associated(CurrentField%ident)) then
		write(*,*)'GetDemandPriceQuantities: DemandPrice%ident is associated'
	        DemandPrice%name = CurrentField%ident%name
	      else
		write(*,*)'GetDemandPriceQuantities: DemandPrice%ident is not associated'
	        DemandPrice%name = CurrentField%identtext
	      end if
	      DemandPrice%arg => CurrentField%arg
	    else if(CurrentLabel.eq.'e') then
! EndowmentPrice
	      EndowmentPrice%type = 'v'
	      if(associated(CurrentField%ident)) then
		write(*,*)'GetDemandPriceQuantities: EndowmentPrice%ident is associated'
	        EndowmentPrice%name = CurrentField%ident%name
	      else
		write(*,*)'GetDemandPriceQuantities: EndowmentPrice%ident is not associated'
	        EndowmentPrice%name = CurrentField%identtext
	      end if
	      EndowmentPrice%arg => CurrentField%arg
	    else if(CurrentLabel.eq.'q') then
! DemandQuantity or EndowmentQuantity
	      if(LastLabel.eq.'d') then
! DemandQuantity
	        DFieldExists = .true .
	        DemandQuantity%type = 'c'
	        if(associated(CurrentField%ident)) then
		  write(*,*)'GetDemandPriceQuantities: DemandQuantity%ident is associated'
	          DemandQuantity%name = CurrentField%ident%name
		else
		  write(*,*)'GetDemandPriceQuantities: DemandQuantity%ident is not associated'
	          DemandQuantity%name = CurrentField%identtext
		end if
	        DemandQuantity%arg => CurrentField%arg
	      else if(LastLabel.eq.'e') then
! EndowmentQuantity
	        EFieldExists = .true .
	        EndowmentQuantity%type = 'c'
	        if(associated(CurrentField%ident)) then
		  write(*,*)'GetDemandPriceQuantities: EndowmentQuantity%ident is associated'
	          EndowmentQuantity%name = CurrentField%ident%name
		else
		  write(*,*)'GetDemandPriceQuantities: EndowmentQuantity%ident is not associated'
	          EndowmentQuantity%name = CurrentField%identtext
		end if
	        EndowmentQuantity%arg => CurrentField%arg
!	      else
!		Error
	      end if
	    end if

	    LastLabel = CurrentLabel
	    CurrentField => CurrentField%next 
	  end do
	  CurrentRecord => CurrentRecord%next
	end do	

	return
	end subroutine GetDemandPriceQuantities

	subroutine GetDemandVariableGP(ThisTopVar, EndowmentPrice, ThisDemandVar)
! Input
	type (GPCoeffVarReal), intent(in) :: ThisTopVar, EndowmentPrice
! Output
	type (GPCoeffVarReal), intent(out) :: ThisDemandVar

	ThisDemandVar%type = 'v'
	ThisDemandVar%name = 'qs_'//trim(ThisTopVar%name)//'_'// &
	    trim(EndowmentPrice%name)
	ThisDemandVar%arg => EndowmentPrice%arg

	return
	end subroutine GetDemandVariableGP

	subroutine WriteDemandUpdate(ThisDemandVar, &
		EndowmentPrice, EndowmentQuantity)
! Input
	type (GPCoeffVarReal), intent(in) :: ThisDemandVar, EndowmentPrice
	type (GPCoeffVarReal), intent(in) :: EndowmentQuantity
! Local
	character (len=512) OutputString, TempString, TempString2

10005	format(a)

! quantifiers and LHS =
	call ArgToAllsGP(EndowmentPrice%arg, TempString)
	call CoeffVarToGP(EndowmentQuantity, TempString2)
	OutputString='Update '//trim(TempString)//' '// &
	  trim(TempString2)//' ='
	write(iotabgp, 10005) trim(OutputString)

! RHS <EndowmentPrice>*<ThisDemandVar>
	call CoeffVarToGP(EndowmentPrice, TempString)
	call CoeffVarToGP(ThisDemandVar, TempString2)
	OutputString = '  '//trim(TempString)//' * '//trim(TempString2)//' ;'
	write(iotabgp, 10005) trim(OutputString)

	return
	end subroutine WriteDemandUpdate

	subroutine WriteDemandMarketClearingEquation(DemandTopVar, DemandPrice)
! Input
	type (GPCoeffVarReal), intent(in) :: DemandTopVar, DemandPrice
! Local
	character (len=512) OutputString, TempString, TempString2
	character (len=512) TempString3, TempString4
	type (GPCoeffVarReal) :: AssociatedProdTopVar
	type (GPCoeffVarReal) :: ThisProdOutputPrice, ThisProdOutputQuantity
	type (GPCoeffVarReal) :: ThisProdTopVar, ThisProdElasticity
	character (len=LengthGPDeclarationName) EquationName
	type (field), pointer :: fnf
	type (record), pointer :: r
	type (function), pointer :: fn
	character (len=128) DemandPriceName, DemandTopVarName
	character (len=128) ProdOutputPriceName, ProdTopVarName
	logical :: AssociatedProdFound
	

10005	format(a)

	EquationName = 'E_'//trim(DemandPrice%name)
	OutputString = 'Equation '//trim(EquationName)//' # '// &
	   'Market clearing for $demand:'//trim(DemandTopVar%name)//' #'
	write(iotabgp, 10005) trim(OutputString)

! Need to find the value associated with DemandPrice
!  Need to look in all $prod blocks to find this.
! Example
!$demand:ra
!  d:pc              q:cons
!  e:pf(f)           q:endow(f)
!$prod:c             s:1.0
!  o:pc              q:cons
!  i:p(s)            q:demand(s)
! Associated value for $demand:ra is "c" since DemandPrice "pc"
!  is output price in the $prod:c block

	DemandPriceName = DemandPrice%name
	DemandTopVarName = DemandTopVar%name
	call lower(DemandPriceName)
	call lower(DemandTopVarName)
	write(*,*)'WriteDemandMarketClearingEquation: DemandTopVarName=', trim(DemandTopVarName)
	AssociatedProdFound = .false.
	fn => functionlist
	do while (associated(fn))
	  r => fn%firstrecord
	  fnf => r%firstfield
	  if (fnf%label.eq.'$prod') then
	    call GetProdPriceQuantity(fn, 'o', ThisProdOutputPrice, ThisProdOutputQuantity)
	    call GetProdTopVariableElast(fn, ThisProdTopVar, ThisProdElasticity)
! Compare them
	    ProdOutputPriceName = ThisProdOutputPrice%name
	    ProdTopVarName = ThisProdTopVar%name
	    call lower(ProdOutputPriceName)
	    call lower(ProdTopVarName)
	    write(*,*)'WriteDemandMarketClearingEquation: ProdOutputPriceName=', trim(ProdOutputPriceName)
	    if(trim(DemandPriceName).eq.trim(ProdOutputPriceName)) then
! Have found the one we want
	      AssociatedProdTopVar = ThisProdTopVar
	      AssociatedProdFound = .true.
	    end if
	  end if
	  if(AssociatedProdFound) then
	    exit  ! from this do while block
	  else
	    fn => fn%next
	  end if
	end do

	if(.not.AssociatedProdFound) then
!	  Error
	  write(*,*)' **ERROR: Associated Prod not found for $demand:'//DemandTopVar%name
	end if	      

! quantifiers and equation
	call ArgToAllsGP(DemandTopVar%arg, TempString)
	call CoeffVarToGP(DemandPrice, TempString2)
	call CoeffVarToGP(AssociatedProdTopVar, TempString3)
	call CoeffVarToGP(DemandTopVar, TempString4)
	OutputString=' '//trim(TempString)//' '// &
	  trim(TempString2)//' + '//trim(TempString3)// &
	  ' = '//trim(TempString4)//' ;'
	write(iotabgp, 10005) trim(OutputString)

	return
	end subroutine WriteDemandMarketClearingEquation

	subroutine WriteDemandIncomeBalanceEquation(ThisDemandTopVar, &
		EndowmentPrice, EndowmentQuantity, ThisDemandVar)
! Input
	type (GPCoeffVarReal), intent(in) :: ThisDemandTopVar, ThisDemandVar
	type (GPCoeffVarReal), intent(in) :: EndowmentPrice, EndowmentQuantity
! Local
	character (len=512) :: OutputString, TempString, TempString2
	character (len=512) :: TempString3, TempString4
	character (len=512) :: TempStringSum, TempStringSumEnd
	character (len=LengthGPDeclarationName) :: EquationName
	integer :: NSum

10005	format(a)

	EquationName = 'E_'//trim(ThisDemandTopVar%name)
	OutputString = 'Equation '//trim(EquationName)//' # '// &
	   'Income balance for $demand:'//trim(ThisDemandTopVar%name)//' #'
	write(iotabgp, 10005) trim(OutputString)

! LHS
!  Sum
	call WriteSumsGP(EndowmentQuantity%arg, TempStringSum, NSum)
! Sum of EndowmentQuantity
	call CoeffVarToGP(EndowmentQuantity, TempString)
	OutputString = '   '//trim(TempStringSum)//' '//trim(TempString)
! End the sums
	call EndSumsGP(NSum, TempStringSumEnd)
	call CoeffVarToGP(ThisDemandTopVar, TempString2)
	OutputString = trim(OutputString)//trim(TempStringSumEnd)// &
	    ' * '//trim(TempString2)//' ='
	write(iotabgp, 10005) trim(OutputString)

! RHS
! Sum of EndowmentQuantity
	call CoeffVarToGP(EndowmentQuantity, TempString)
	OutputString = '    '//trim(TempStringSum)//' '//trim(TempString)
	call CoeffVarToGP(EndowmentPrice, TempString2)
	call CoeffVarToGP(ThisDemandVar, TempString3)
	OutputString = trim(OutputString)//' * ['// &
	  trim(TempString2)//' + '//trim(TempString3)//']'
! End the sums
	OutputString = trim(OutputString)//' '//trim(TempStringSumEnd)//' ;'
	write(iotabgp, 10005) trim(OutputString)
	
	return
	end subroutine WriteDemandIncomeBalanceEquation

	subroutine WriteArguments(FirstArg, OutputText, LenOutput)
! Input
	type (domain), pointer :: FirstArg
! Output
	character (len=255) OutputText
	integer LenOutput
! Local
	type (domain), pointer :: CurrentArg
	type (symbol), pointer :: CurrentSet

10017   format(2x,3a)
	OutputText = ''
	LenOutput = 0	
	CurrentArg => FirstArg	
	do while(associated(CurrentArg))
	  CurrentSet => CurrentArg%s
	  OutputText(LenOutput+1:) = trim(CurrentSet%name)//' '
	  LenOutput = LenOutput + LEN(trim(CurrentSet%name))+1
	  CurrentArg => CurrentArg%next
	end do
!	write(*,10017) 'Arguments: ', OutputText(1:LenOutput) 

	return
	end subroutine WriteArguments 
	
	subroutine WriteRecord(DescText, ThisRecord)
! Input
	character(len=*), intent(in) :: DescText
	type (record), pointer :: ThisRecord
! Local
	type (record), pointer :: CurrentRecord
	type (field), pointer :: CurrentField

	CurrentRecord => ThisRecord
	do while(associated(CurrentRecord))
	  write(*,*)'WriteRecord - a new record'
	  CurrentField => CurrentRecord%firstfield
	  do while(associated(CurrentField))
	    call WriteField('Next field in record', CurrentField)
	    CurrentField => CurrentField%next 
	  end do
	  CurrentRecord => CurrentRecord%next
	end do	

	return
	end subroutine WriteRecord
	
	subroutine WriteField(DescText, ThisField)
! Input
	character(len=*), intent(in) :: DescText
	type (field), pointer :: ThisField
! Local
	character*256 OutputText
	integer LenOutput

10017   format(2x,3a)

	if (associated(ThisField)) then
	  write(*,10017) DescText, '%data: ', trim(ThisField%data)
	  if(associated(ThisField%ident)) then
	    write(*,10017) DescText, '%ident%name: ', trim(ThisField%ident%name)
	  end if
	  call WriteArguments(ThisField%arg, OutputText, LenOutput) 
	  write(*,10017) DescText, ' Arguments: ', OutputText(1:LenOutput)
	end if

	return
	end subroutine WriteField

	subroutine GPSet2Index(CurrentSet, CurrentIndex)
! Input
	type (symbol), pointer :: CurrentSet
! Output
	character*(*) CurrentIndex
!
	CurrentIndex = trim(CurrentSet%name)//'1'

	return
	end subroutine GPSet2Index

! Next replaced by a test running through GPDeclarationList
!	logical function HasBeenDeclaredGP(ThisSymbol)
! Tells if ThisSymbol has already been declared
! Input
!	type (symbol), pointer :: ThisSymbol
! Local
!	type (symbol), pointer :: TestSymbol
!	character(len=31) ThisName, TestName

!	HasBeenDeclaredGP = .false.

!	if(.not.associated(ThisSymbol)) then
!	  write(*,*) 'HasBeenDeclaredGP - symbol is not associated'
!	  return
!	end if
!	ThisName = trim(ThisSymbol%name)
!	call lower(ThisName)

! Run through all the symbols (they are in idlist) 
!  to find this one and check if it is declared
!	TestSymbol => idlist ;
!	do while(associated(TestSymbol))
!	  TestName = trim(TestSymbol%name)
!	  call lower(TestName)
! Are the two names the same ?
!	  if (TestName.eq.ThisName) then
!	    if(TestSymbol%DeclaredGP) then
!	      HasBeenDeclaredGP = .true.
!	      write(*,*)'HasBeenDeclaredGP:',trim(ThisName),' has been declared'
!	      Exit
!	    else
!	      write(*,*)'HasBeenDeclaredGP:',trim(ThisName),' says not declared'
!	    end if
!	  end if
!	  TestSymbol => TestSymbol%next
!	end do 

!	return
!	end function HasBeenDeclaredGP

	subroutine MarkAsDeclaredGP(ThisSymbol)
! Indicate that ThisSymbol has been declared
! Input and output
	type (symbol), pointer :: ThisSymbol
! Local
	type (symbol), pointer :: TestSymbol
	character(len=31) ThisName, TestName

	ThisName = trim(ThisSymbol%name)
	call lower(ThisName)

! Run through all the symbols (they are in idlist) 
!  to find this one and then mark it as declared
	TestSymbol => idlist ;
	do while(associated(TestSymbol))
	  TestName = trim(TestSymbol%name)
	  call lower(TestName)
! Are the two names the same ?
	  if (TestName.eq.ThisName) then
	    TestSymbol%DeclaredGp = .true .
	    write(*,*)'MarkAsDeclaredGP:',trim(ThisName),' now declared'
	  end if
	  TestSymbol => TestSymbol%next
	end do 

! Temporary
! Run through all the symbols again 
!  to find this one and then check it is shown as declared
	TestSymbol => idlist ;
	do while(associated(TestSymbol))
	  TestName = trim(TestSymbol%name)
	  call lower(TestName)
! Are the two names the same ?
	  if (TestName.eq.ThisName) then
	    if(TestSymbol%DeclaredGp) then
	      write(*,*)'MarkAsDeclaredGP:',trim(ThisName),' shown as declared'
	    else
	      write(*,*)'MarkAsDeclaredGP:',trim(ThisName),' NOT shown as declared'
	    end if
	  end if
	  TestSymbol => TestSymbol%next
	end do 

	return
	end subroutine MarkAsDeclaredGP

	subroutine ArgToAllsGP(TheseArgs, OutputString)
! Input
	type (domain), pointer :: TheseArgs
! Output
	character (len=*), intent(out) :: OutputString
! Local
	integer LenOutputString
	type (domain), pointer :: CurrentArgument
	type (symbol), pointer :: CurrentSet
	character (len=128) CurrentIndex, CurrentAll
	integer LengthCurrentAll

	OutputString = ''
	LenOutputString = 0 

	CurrentArgument => TheseArgs
! The ALLs
	do while(associated(CurrentArgument))
	  CurrentSet => CurrentArgument%s
	  call GPSet2Index(CurrentSet, CurrentIndex)
	  CurrentAll = '(All,'//trim(CurrentIndex)//','//trim(CurrentSet%name)//')'
	  LengthCurrentAll = 7 + LEN(trim(CurrentIndex)) + LEN(trim(CurrentSet%name)) 
	  OutputString = OutputString(1:LenOutputString) // CurrentAll
	  LenOutputString = LenOutputString + LEN(trim(CurrentAll))
	  CurrentArgument => CurrentArgument%next
	end do

	return
	end subroutine ArgToAllsGP

	subroutine CoeffVarToGP(ThisCoeffVar, OutputString)
! Writes coeff/var followed by its arguments
! Input
	type (GPCoeffVarReal) :: ThisCoeffVar
! Output
	character (len=*), intent(out) :: OutputString
! Local
	type (domain), pointer :: TheseArgs
	character (len=512) :: ArgString

	OutputString = ThisCoeffVar%name
	if(ThisCoeffVar%type.eq.'c') then
! Upper case for Coefficients
	  call upper(OutputString)
	end if
	TheseArgs => ThisCoeffVar%arg
	call ArgToGP(TheseArgs, ArgString)
	if(ArgString.ne.'') then
	  OutputString = trim(OutputString)//'('//trim(ArgString)//')'
	end if
	OutputString = trim (OutputString)

	return
	end subroutine CoeffVarToGP
	
	subroutine ArgToGP(TheseArgs, OutputString)
! Write TheseArgs in form suitable for GEMPACK. eg "s1,f1"
! No start "(" or end ")" is added here
! Input
	type (domain), pointer :: TheseArgs
! Output
	character (len=*), intent(out) :: OutputString
! Local
	integer NumArg
	type (domain), pointer :: CurrentArgument
	type (symbol), pointer :: CurrentSet
	character (len=128) CurrentIndex, CurrentAll

	OutputString = ''
	NumArg = 0 

	CurrentArgument => TheseArgs
	do while(associated(CurrentArgument))
	  if(NumArg.gt.0) then
! comma to separate arguments
	    OutputString = trim(OutputString)//','
	  end if
	  CurrentSet => CurrentArgument%s
	  call GPSet2Index(CurrentSet, CurrentIndex)
	  OutputString = trim(OutputString)//trim(CurrentIndex)
	  NumArg = NumArg + 1
	  CurrentArgument => CurrentArgument%next
	end do

	return
	end subroutine ArgToGP

	subroutine DeclareFieldGP(DeclarationType, ThisField)
! Input
	character(*), intent(in) :: DeclarationType  ! 'variable', 'coefficient' etc
	type (field), pointer :: ThisField
! Local
	character(len=512) OutputString, OutputStringAlls
	character(len=512) OutputStringArgs, CoeffVarName
	integer NumArg, karg, NumArg2, NumArg3
	type (symbol), pointer :: CurrentSet
	character (len=128) CurrentIndex, CurrentAll
	integer LengthCurrentAll
	type (domain), pointer :: CurrentArgument
	character (len=LengthGPDeclarationName) TypeName
	character (len=3) TypeStart
	character (len=2) ThisType
	logical IsCoeff

10005   format(a)
10017   format(2x,3a)

! If not associated, exit
	if(.not.associated(ThisField)) Return
! If identifier is not associated, exit
	if(.not.associated(ThisField%ident)) Return

!  Set up TypeName and IsCoeff and ThisType
	TypeStart = DeclarationType
	call lower(TypeStart)
	IsCoeff = .false.
	if(TypeStart.eq.'coe') then
	  TypeName = 'Coefficient'
	  IsCoeff = .true.
	else if(TypeStart.eq.'var') then
	  TypeName = 'Variable'
	end if
	if (IsCoeff) then
	  ThisType = 'cs'
	else
	  ThisType = 'vc'
	end if

! If already declared, exit
! WAS	if (HasBeenDeclaredGP(ThisField%ident)) Return
	if (HasBeenDeclaredGP(ThisType,ThisField%ident%name)) Return

	OutputString = ''

	OutputString=trim(TypeName)//' '
	
	call ArgToAllsGP(ThisField%arg, OutputStringAlls)
	OutputStringAlls = trim(OutputStringAlls)
	if(OutputStringAlls.ne.'') then
	  OutputString = trim(OutputString)//OutputStringAlls
	end if

! The name
	CoeffVarName = ThisField%ident%name
! Coefficients in upper case
	if(IsCoeff) then
	  call upper(CoeffVarName)
	end if
	write(*,*)'ThisFile%ident%name): ', ThisField%ident%name
	OutputString = trim(OutputString)//' '//trim(CoeffVarName)
! The arguments
	if(OutputStringAlls.ne.'') then
! Opening (
	  OutputString = trim(OutputString)//'('
	end if
	call ArgToGP(ThisField%arg, OutputStringArgs)
	CurrentArgument => ThisField%arg
	OutputString = trim(OutputString)//trim(OutputStringArgs)
	if(OutputStringAlls.ne.'') then
! Closing (
	  OutputString = trim(OutputString)//')'
	end if

! The semicolon
	OutputString = trim(OutputString)//' ;'
! Write the declaration
	write(*,10017) 'Declaration: ', trim(OutputString)
	write(iotabgp, 10005) trim(OutputString)
! Mark this symbol as declared
	call MarkAsDeclaredGP(ThisField%ident)
! Mark it in GPDeclarations
	call AddGPDeclaration(ThisType, trim(ThisField%ident%name))
 		
	return
	end subroutine DeclareFieldGP


	subroutine WriteNest(DescText, ThisNest)
	character*(*) DescText
	type (nest), pointer :: ThisNest
! Local
	type (netputlist), pointer :: elements
	type (netput), pointer :: ThisNetPut
	type (field), pointer :: IOField, qref, pref

10015   format(1x,a)

	elements => ThisNest%elements
	do while (associated(elements)) 
	  ThisNetPut => elements%np
	  IOField => ThisNetPut%f   ! I: or O: field
	  if (ThisNetPut%output) then
	    write(*,10015) ' An output'
	  else
	    write(*,10015) ' An input'
	  end if
	  qref => ThisNetPut%qref
	  pref => ThisNetPut%pref

	  call WriteField('IOField', IOField)
	  if(associated(IOField)) then
!	    call DeclareFieldGP('Variable', IOField)
	  end if

	  call WriteField('QREF', qref)
	  if(associated(qref)) then
!	    call DeclareFieldGP('Coefficient', qref)
	  end if
	  call WriteField('PREF', pref)
	  if(associated(pref)) then
!	    call DeclareFieldGP('Coefficient', pref)
	  end if
	  elements => elements%next
	end do
	
	return
	end subroutine WriteNest

!!	Allocate a new GPDeclaration and assign default values
!!	for all the characteristics

	subroutine NewGPDeclaration(ThisGPDeclaration)
	type (GPDeclaration), pointer :: ThisGPDeclaration
	nullify(ThisGPDeclaration)
	allocate(ThisGPDeclaration)
	ThisGPDeclaration%type = ''  ! impossible type
	ThisGPDeclaration%name = ''  ! impossible name
	nullify(ThisGPDeclaration%next)
	return
	end subroutine NewGPDeclaration

!! Introduce a new GPDeclaration in the master list.

        subroutine AddGPDeclaration(type, name)      

        character (len=*), intent(in) :: type
        character (len=*), intent(in) :: name

        type (GPDeclaration), pointer :: ThisGPDeclaration

	write(*,*)'AddGPDeclaration: type=', type

        if (.not.associated(GPDeclarationList)) then
	  write(*,*)'AddGPDeclaration: not associated'
	  call NewGPDeclaration(GPDeclarationList)
          ThisGPDeclaration => GPDeclarationList
        else
	  write(*,*)'AddGPDeclaration: is associated'
	  call NewGPDeclaration(GPDeclarationLast%next)
          ThisGPDeclaration => GPDeclarationLast%next
        end if
        ThisGPDeclaration%type = type
        ThisGPDeclaration%name = name
        nullify(ThisGPDeclaration%next)
        GPDeclarationLast => ThisGPDeclaration
        return
        end subroutine AddGPDeclaration

	logical function HasBeenDeclaredGP(type, name)
! Tells if this has already been declared
! Input
        character (len=*), intent(in) :: type
        character (len=*), intent(in) :: name
! Local
	type (GPDeclaration), pointer :: CurrentGPDeclaration
	character(len=2) CurrentType, TestType
	character(len=LengthGPDeclarationName) CurrentName, TestName

	HasBeenDeclaredGP = .false.

	TestType = type
	call lower(TestType)
	TestName = trim(name)
	call lower(TestName)

! Run throught the list of declarations seeking a match
	CurrentGPDeclaration => GPDeclarationList
	do while(associated(CurrentGPDeclaration))
	  CurrentType = CurrentGPDeclaration%type
	  write(*,*)'HasBeenDeclaredGP: CurrentType=', CurrentType
	  call lower(CurrentType)
	  CurrentName = CurrentGPDeclaration%name
	  CurrentName = trim(CurrentName)
	  write(*,*)'HasBeenDeclaredGP: CurrentName=', CurrentName
	  call lower(CurrentName)
	  if(CurrentName.eq.TestName) then
	    if(CurrentType.eq.TestType) then
	      HasBeenDeclaredGP = .true.
	      return
	    end if
	  end if
	  CurrentGPDeclaration => CurrentGPDeclaration%next
	end do	
	
	return
	end function HasBeenDeclaredGP
	

end module mpsge

